@@ -4,13 +4,14 @@ for all indexes.
 
 """
 from __future__ import unicode_literals
+from io import BytesIO
 import hashlib
 import mimetypes
 from wsgiref.handlers import format_date_time
-import os
 import py
 import re
 import sys
+from devpi_common.metadata import splitbasename
 from devpi_common.types import cached_property, parse_hash_spec
 from .log import threadlog
 
@@ -48,7 +49,10 @@ class FileStore:
         self.rel_storedir = "+files"
         self.storedir = self.keyfs.basedir.join(self.rel_storedir)
 
-    def maplink(self, link, user, index):
+    def maplink(self, link, user, index, project):
+        parts = link.torelpath().split("/")
+        assert parts
+        basename = unquote(parts[-1])
         if link.hash_spec:
             # we can only create 32K entries per directory
             # so let's take the first 3 bytes which gives
@@ -58,14 +62,12 @@ class FileStore:
                                        hashdir_a=a, hashdir_b=b,
                                        filename=link.basename)
         else:
-            parts = link.torelpath().split("/")
-            assert parts
             dirname = "_".join(parts[:-1])
             dirname = re.sub('[^a-zA-Z0-9_.-]', '_', dirname)
             key = self.keyfs.PYPIFILE_NOMD5(
                 user=user, index=index,
                 dirname=unquote(dirname),
-                basename=unquote(parts[-1]))
+                basename=basename)
         entry = FileEntry(self.xom, key, readonly=False)
         entry.url = link.geturl_nofragment().url
         entry.eggfragment = link.eggfragment
@@ -80,17 +82,38 @@ class FileStore:
                 threadlog.error(err)
                 entry.file_delete()
         entry.hash_spec = unicode_if_bytes(link.hash_spec)
+        entry.project = project
+        version = None
+        if link.eggfragment:
+            version = link.eggfragment[len(project) + 1:]
+        else:
+            try:
+                (projectname, version, ext) = splitbasename(basename)
+            except ValueError:
+                pass
+        # only store version on entry if we can determine it
+        # since version is a meta property of FileEntry, it will return None
+        # if not set, if we set it explicitly, it would waste space in the
+        # database
+        if version is not None:
+            entry.version = version
         return entry
 
-    def get_file_entry(self, relpath, readonly=True):
+    def get_key_from_relpath(self, relpath):
         try:
             key = self.keyfs.tx.derive_key(relpath)
         except KeyError:
             return None
-        return FileEntry(self.xom, key, readonly=readonly)
+        return key
 
-    def get_file_entry_raw(self, key, meta):
-        return FileEntry(self.xom, key, meta=meta)
+    def get_file_entry(self, relpath, readonly=True):
+        key = self.get_key_from_relpath(relpath)
+        if key is None:
+            return None
+        return self.get_file_entry_from_key(key, readonly=readonly)
+
+    def get_file_entry_from_key(self, key, meta=_nodefault, readonly=True):
+        return FileEntry(self.xom, key, meta=meta, readonly=readonly)
 
     def store(self, user, index, basename, file_content, dir_hash_spec=None):
         if dir_hash_spec is None:
@@ -137,9 +160,9 @@ class FileEntry(object):
         self.relpath = key.relpath
         self.basename = self.relpath.split("/")[-1]
         self.readonly = readonly
-        self._storepath = os.path.join(
+        self._storepath = "/".join((
             self.xom.filestore.rel_storedir,
-            str(self.relpath))
+            str(self.relpath)))
         if meta is not _nodefault:
             self.meta = meta or {}
 
@@ -212,6 +235,11 @@ class FileEntry(object):
         # end up only committing file content without any keys
         # changed which will not replay correctly at a replica.
         self.key.set(self.meta)
+        stage = self.xom.model.getstage(
+            self.key.params['user'],
+            self.key.params['index'])
+        if self.project:
+            stage.add_project_name(self.project)
 
     def gethttpheaders(self):
         assert self.file_exists()
@@ -220,6 +248,7 @@ class FileEntry(object):
         m = mimetypes.guess_type(self.basename)[0]
         headers[str("content-type")] = str(m)
         headers[str("content-length")] = str(self.file_size())
+        headers[str("cache-control")] = str("max-age=365000000, immutable, public")
         return headers
 
     def __eq__(self, other):
@@ -250,6 +279,9 @@ class FileEntry(object):
             headers[str("content-length")] = str(r.headers["content-length"])
         return headers
 
+    def has_existing_metadata(self):
+        return self.hash_spec and self.last_modified
+
     def iter_cache_remote_file(self):
         # we get and cache the file and some http headers from remote
         r = self.xom.httpget(self.url, allow_redirects=True)
@@ -263,14 +295,16 @@ class FileEntry(object):
 
         yield self._headers_from_response(r)
 
-        content = b''
+        content = BytesIO()
         while 1:
             data = r.raw.read(10240)
             if not data:
                 break
-            content = content + data
+            content.write(data)
             yield data
 
+        content = content.getvalue()
+
         filesize = len(content)
         if content_size and int(content_size) != filesize:
             err = ValueError(
@@ -291,11 +325,20 @@ class FileEntry(object):
             # when streaming we won't be in a transaction anymore, so we need
             # to open a new one below
             tx = None
-        if tx is not None:
-            self.file_set_content(content, r.headers.get("last-modified", None))
-        else:
-            with self.key.keyfs.transaction(write=True):
+        if not self.has_existing_metadata():
+            if tx is not None:
                 self.file_set_content(content, r.headers.get("last-modified", None))
+            else:
+                with self.key.keyfs.transaction(write=True):
+                    self.file_set_content(content, r.headers.get("last-modified", None))
+        else:
+            # the file was downloaded before but locally removed, so put
+            # it back in place without creating a new serial
+            # we need a direct write connection to use the io_file_* methods
+            with self.key.keyfs._storage.get_connection(write=True) as conn:
+                conn.io_file_set(self._storepath, content)
+                log.debug("put missing file back into place: %s", self._storepath)
+                conn.commit_files_without_increasing_serial()
 
     def iter_remote_file_replica(self):
         from .replica import H_REPLICA_FILEREPL, ReplicationErrors
@@ -323,14 +366,16 @@ class FileEntry(object):
 
         yield self._headers_from_response(r)
 
-        content = b''
+        content = BytesIO()
         while 1:
             data = r.raw.read(10240)
             if not data:
                 break
-            content = content + data
+            content.write(data)
             yield data
 
+        content = content.getvalue()
+
         err = self.check_checksum(content)
         if err:
             # the file we got is different, so we fail
@@ -350,7 +395,8 @@ class FileEntry(object):
             # we need a direct write connection to use the io_file_* methods
             with self.key.keyfs._storage.get_connection(write=True) as conn:
                 conn.io_file_set(self._storepath, content)
-                conn.commit()
+                log.debug("put missing file back into place: %s", self._storepath)
+                conn.commit_files_without_increasing_serial()
         # in case there were errors before, we can now remove them
         replication_errors.remove(self)
 
