@@ -15,6 +15,25 @@ from .main import fatal
 from .readonly import get_mutable_deepcopy, ReadonlyView
 
 
+def has_users_or_stages(xom):
+    userlist = xom.model.get_userlist()
+    if len(userlist) == 0:
+        # no data at all
+        return False
+    if len(userlist) == 1:
+        # we got one user, check to see if it's the default root user
+        user, = userlist
+        if user.name == "root":
+            rootindexes = list(user.get().get("indexes", []))
+            if not rootindexes:
+                # the root user has no indexes
+                return False
+            # it's fine if only the default pypi index exists
+            if rootindexes == ["pypi"]:
+                return False
+    return True
+
+
 def do_export(path, xom):
     path = py.path.local(path)
     tw = py.io.TerminalWriter()
@@ -35,8 +54,8 @@ def do_import(path, xom):
     if not path.check():
         fatal("path for importing not found: %s" %(path))
 
-    with xom.keyfs.transaction():
-        if not xom.model.is_empty():
+    with xom.keyfs.transaction(write=False):
+        if has_users_or_stages(xom):
             fatal("serverdir must not contain users or stages: %s" %
                   xom.config.serverdir)
     importer = Importer(tw, xom)
@@ -309,7 +328,7 @@ class Importer:
                     if stage is not None:
                         continue
                 import_index = self.import_indexes[stagename]
-                indexconfig = import_index["indexconfig"]
+                indexconfig = dict(import_index["indexconfig"])
                 if 'uploadtrigger_jenkins' in indexconfig:
                     if not indexconfig['uploadtrigger_jenkins']:
                         # remove if not set, so if the trigger was never
@@ -322,7 +341,13 @@ class Importer:
                         indexconfig['mirror_whitelist'] = whitelist
                 user, index = stagename.split("/")
                 user = self.xom.model.get_user(user)
+                # due to possible circles we create without bases first
+                bases = indexconfig.pop('bases')
                 stage = user.create_stage(index, **indexconfig)
+                if "bases" in import_index["indexconfig"]:
+                    indexconfig = stage.ixconfig
+                    indexconfig["bases"] = bases
+                    stage.modify(**indexconfig)
                 stages.append(stage)
         del tree
 
@@ -432,7 +457,10 @@ class IndexTree:
         self.name2bases = {}
 
     def add(self, name, bases=None):
-        self.name2bases[name] = bases or []
+        bases = list(bases or [])
+        while name in bases:
+            bases.remove(name)
+        self.name2bases[name] = bases
         if not bases:
             self.name2children.setdefault(None, []).append(name)
         else:
