@@ -338,6 +338,13 @@ class TypedKey:
     def get(self, readonly=True):
         return self.keyfs.tx.get(self, readonly=readonly)
 
+    @property
+    def last_serial(self):
+        try:
+            return self.keyfs.tx.last_serial(self)
+        except KeyError:
+            return None
+
     def is_dirty(self):
         return self.keyfs.tx.is_dirty(self)
 
@@ -374,8 +381,9 @@ class Transaction(object):
         self.cache = {}
         self.dirty = set()
         self.closed = False
+        self.doomed = False
 
-    def get_value_at(self, typedkey, at_serial):
+    def _get_value_at(self, typedkey, at_serial):
         relpath = typedkey.relpath
         keyname, last_serial = self.conn.db_read_typedkey(relpath)
         while last_serial >= 0:
@@ -386,13 +394,21 @@ class Transaction(object):
                 last_serial = back_serial
                 continue
             if val is not None:
-                return val
+                return (last_serial, val)
             raise KeyError(relpath)  # was deleted
 
         # we could not find any change below at_serial which means
         # the key didn't exist at that point in time
         raise KeyError(relpath)
 
+    def get_value_at(self, typedkey, at_serial):
+        (last_serial, val) = self._get_value_at(typedkey, at_serial)
+        return val
+
+    def last_serial(self, typedkey):
+        (last_serial, val) = self._get_value_at(typedkey, self.at_serial)
+        return last_serial
+
     def derive_key(self, relpath):
         """ return key instance for a given key path."""
         try:
@@ -449,7 +465,6 @@ class Transaction(object):
         else:
             return True
 
-
     def delete(self, typedkey):
         if not self.write:
             raise self.keyfs.ReadOnly()
@@ -467,6 +482,9 @@ class Transaction(object):
         self.dirty.add(typedkey)
 
     def commit(self):
+        if self.doomed:
+            threadlog.debug("closing doomed transaction")
+            return self._close()
         if not self.write:
             return self._close()
         if not self.dirty and not self.conn.dirty_files:
@@ -511,6 +529,10 @@ class Transaction(object):
         newtx = self.__class__(self.keyfs, write=write)
         self.__dict__ = newtx.__dict__
 
+    def doom(self):
+        """ mark as doomed to automatically rollback any changes """
+        self.doomed = True
+
 
 def check_unicode_keys(d):
     for key, val in d.items():
