@@ -1,4 +1,5 @@
 from devpi_common.types import cached_property
+from .config import hookimpl
 from .fileutil import dumps, loads
 from .log import threadlog, thread_push_log, thread_pop_log
 from .readonly import ReadonlyView
@@ -96,6 +97,7 @@ class Connection(BaseConnection):
         q = "INSERT OR REPLACE INTO files (path, size, data) VALUES (?, ?, ?)"
         c.execute(q, (path, len(content), sqlite3.Binary(content)))
         c.close()
+        self.dirty_files[path] = content
 
     def io_file_open(self, path):
         return py.io.BytesIO(self.io_file_get(path))
@@ -127,10 +129,14 @@ class Connection(BaseConnection):
         q = "DELETE FROM files WHERE path = ?"
         c.execute(q, (path,))
         c.close()
+        self.dirty_files[path] = None
 
     def write_transaction(self):
         return Writer(self.storage, self)
 
+    def commit_files_without_increasing_serial(self):
+        self.commit()
+
 
 class BaseStorage:
     def __init__(self, basedir, notify_on_commit, cache_size):
@@ -166,14 +172,19 @@ class BaseStorage:
         except TypeError as e:
             if e.args and 'uri' in e.args[0] and 'keyword argument' in e.args[0]:
                 threadlog.warn(
-                    "Can't open sqlite3 db with uri keyword. Python 3.4 is "
-                    "the first version to support it.")
+                    "The uri keyword for 'sqlite3.connect' isn't supported by "
+                    "this Python version.")
             else:
                 raise
+        except sqlite3.OperationalError as e:
+            threadlog.warn("%s" % e)
+            threadlog.warn(
+                "The installed version of sqlite3 doesn't seem to support "
+                "the uri keyword for 'sqlite3.connect'.")
         except sqlite3.NotSupportedError:
             threadlog.warn(
-                "Can't open sqlite3 db with uri. The installed version of "
-                "sqlite doesn't support it.")
+                "The installed version of sqlite3 doesn't support the uri "
+                "keyword for 'sqlite3.connect'.")
         try:
             # sqlite3 might be compiled with default URI support
             conn = self._get_sqlconn_uri(uri)
@@ -184,7 +195,7 @@ class BaseStorage:
             # log the error and switch to using the path
             threadlog.warn("%s" % e)
             threadlog.warn(
-                "Can't open sqlite3 db with options in URI. There is a "
+                "Opening the sqlite3 db without options in URI. There is a "
                 "higher possibility of read/write conflicts between "
                 "threads, causing slowdowns due to retries.")
             conn = self._get_sqlconn_path(uri)
@@ -255,6 +266,7 @@ class Storage(BaseStorage):
             conn.commit()
 
 
+@hookimpl
 def devpiserver_storage_backend(settings):
     return dict(
         storage=Storage,
