@@ -2,7 +2,7 @@ import threading
 import logging
 import logging.config
 import contextlib
-import yaml
+import json
 
 
 threadlocal = threading.local()
@@ -24,7 +24,11 @@ def configure_logging(config_args):
 
     if config_args.logger_cfg:
         with open(config_args.logger_cfg, 'rt') as f:
-            logger_cfg = yaml.load(f.read())
+            if config_args.logger_cfg.endswith(".json"):
+                logger_cfg = json.loads(f.read())
+            else:
+                import yaml
+                logger_cfg = yaml.load(f.read())
         logging.config.dictConfig(logger_cfg)
 
 
@@ -49,7 +53,7 @@ class TagLogger:
         self._logout.info(self._prefix + msg, *args)
 
     def warn(self, msg, *args):
-        self._logout.warn(self._prefix + msg, *args)
+        self._logout.warning(self._prefix + msg, *args)
 
     def error(self, msg, *args):
         self._logout.error(self._prefix + msg, *args)
