@@ -1,20 +1,21 @@
 # PYTHON_ARGCOMPLETE_OK
 """
 a WSGI server to serve PyPI compatible indexes and a full
-recursive cache of pypi.python.org packages.
+recursive cache of pypi.org packages.
 """
 from __future__ import unicode_literals
-
+import inspect
 import os, sys
 import py
 
-from requests import Response
+from requests import Response, exceptions
 from devpi_common.metadata import Version
 from devpi_common.types import cached_property
 from devpi_common.request import new_requests_session
 from .config import parseoptions, get_pluginmanager
 from .log import configure_logging, threadlog
 from .model import BaseStage
+from .views import apireturn
 from . import mythread
 from . import __version__ as server_version
 
@@ -76,7 +77,7 @@ def _main(pluginmanager, argv=None):
 
     if args.genconfig:
         from devpi_server.genconfig import genconfig
-        return genconfig(config)
+        return genconfig(config, argv[1:])
 
     configure_logging(config.args)
 
@@ -102,7 +103,7 @@ def _main(pluginmanager, argv=None):
         tw = py.io.TerminalWriter()
         bgserver = BackgroundServer(tw, xprocdir)
         if args.start:
-            return bgserver.start(args)
+            return bgserver.start(args, argv[1:])
         elif args.stop:
             return bgserver.stop()
         elif args.log:
@@ -136,6 +137,7 @@ def wsgi_run(xom, app):
     host = xom.config.args.host
     port = xom.config.args.port
     threads = xom.config.args.threads
+    max_body = xom.config.args.max_request_body_size
     log = xom.log
     log.info("devpi-server version: %s", server_version)
     log.info("serverdir: %s" % xom.config.serverdir)
@@ -151,12 +153,23 @@ def wsgi_run(xom, app):
         app = make_eval_exception(app, {})
     try:
         log.info("Hit Ctrl-C to quit.")
-        serve(app, host=host, port=port, threads=threads)
+        serve(app, host=host, port=port, threads=threads, max_request_body_size=max_body)
     except KeyboardInterrupt:
         pass
     return 0
 
 
+def get_caller_location():
+    frame = inspect.currentframe()
+    if frame is None:
+        return 'unknown (no current frame)'
+    caller_frame = frame.f_back.f_back
+    return "%s:%s::%s" % (
+        caller_frame.f_code.co_filename,
+        caller_frame.f_lineno,
+        caller_frame.f_code.co_name)
+
+
 class XOM:
     class Exiting(SystemExit):
         pass
@@ -202,6 +215,15 @@ class XOM:
         """ delete all singletones for the given indexpath """
         self._stagecache.pop(indexpath, None)
 
+    @cached_property
+    def supported_features(self):
+        results = set((
+            'server-keyvalue-parsing',
+        ))
+        for features in self.config.hook.devpiserver_get_features():
+            results.update(features)
+        return tuple(sorted(results))
+
     @cached_property
     def model(self):
         """ root model object. """
@@ -271,7 +293,7 @@ class XOM:
     def _httpsession(self):
         return self.new_http_session("server", max_retries=self.config.args.replica_max_retries)
 
-    def httpget(self, url, allow_redirects, timeout=30, extra_headers=None):
+    def httpget(self, url, allow_redirects, timeout=None, extra_headers=None):
         if self.config.args.offline_mode:
             resp = Response()
             resp.status_code = 503  # service unavailable
@@ -284,11 +306,20 @@ class XOM:
                         url, stream=True,
                         allow_redirects=allow_redirects,
                         headers=headers,
-                        timeout=timeout)
+                        timeout=timeout or self.config.args.request_timeout)
             return resp
+        except OSError:
+            location = get_caller_location()
+            threadlog.warn("OS error during httpget of %s at %s", url, location)
+            return FatalResponse(repr(sys.exc_info()[1]))
+        except exceptions.ConnectionError:
+            location = get_caller_location()
+            threadlog.warn("Connection error during httpget of %s at %s", url, location)
+            return FatalResponse(repr(sys.exc_info()[1]))
         except self._httpsession.Errors:
-            threadlog.exception("Error during httpget of %s", url)
-            return FatalResponse(sys.exc_info())
+            location = get_caller_location()
+            threadlog.warn("HTTPError during httpget of %s at %s", url, location)
+            return FatalResponse(repr(sys.exc_info()[1]))
 
     def create_app(self):
         from devpi_server.view_auth import DevpiAuthenticationPolicy
@@ -328,9 +359,12 @@ class XOM:
         pyramid_config.add_route("/+login", "/+login", accept="application/json")
         pyramid_config.add_route("/{user}/{index}/+e/{relpath:.*}", "/{user}/{index}/+e/{relpath:.*}")
         pyramid_config.add_route("/{user}/{index}/+f/{relpath:.*}", "/{user}/{index}/+f/{relpath:.*}")
+        pyramid_config.add_route("/{user}/{index}/+simple", "/{user}/{index}/+simple")
         pyramid_config.add_route("/{user}/{index}/+simple/", "/{user}/{index}/+simple/")
         pyramid_config.add_route("/{user}/{index}/+simple/{project}",
-                                 "/{user}/{index}/+simple/{project:[^/]+/?}")
+                                 "/{user}/{index}/+simple/{project}")
+        pyramid_config.add_route("/{user}/{index}/+simple/{project}/",
+                                 "/{user}/{index}/+simple/{project}/")
         pyramid_config.add_route("/{user}/{index}/+simple/{project}/refresh",
                                  "/{user}/{index}/+simple/{project}/refresh")
         pyramid_config.add_route("/{user}/{index}/{project}/{version}",
@@ -363,6 +397,7 @@ class XOM:
         pyramid_config.add_request_method(get_remote_ip)
         pyramid_config.add_request_method(stage_url)
         pyramid_config.add_request_method(simpleindex_url)
+        pyramid_config.add_request_method(apifatal)
 
         # overwrite route_url method with our own
         pyramid_config.add_request_method(route_url)
@@ -391,12 +426,8 @@ class XOM:
 class FatalResponse:
     status_code = -1
 
-    def __init__(self, excinfo=None):
-        self.reason = repr(excinfo[1])
-
-    @property
-    def status(self):
-        return "%s %s" % (self.status_code, self.reason)
+    def __init__(self, reason):
+        self.reason = reason
 
     def __repr__(self):
         return "%s(%r)" % (self.__class__.__name__, self.reason)
@@ -431,6 +462,11 @@ def simpleindex_url(request, *args):
             "/{user}/{index}/+simple/", user=stage.username, index=stage.index)
 
 
+def apifatal(request, *args, **kwargs):
+    request.registry["xom"].keyfs.tx.doom()
+    apireturn(*args, **kwargs)
+
+
 def set_default_indexes(model):
     root_user = model.get_user("root")
     if not root_user:
@@ -446,8 +482,8 @@ def set_default_indexes(model):
 _pypi_ixconfig_default = {
     "type": "mirror", "volatile": False,
     "title": "PyPI",
-    "mirror_url": "https://pypi.python.org/simple/",
-    "mirror_web_url_fmt": "https://pypi.python.org/pypi/{name}"}
+    "mirror_url": "https://pypi.org/simple/",
+    "mirror_web_url_fmt": "https://pypi.org/project/{name}/"}
 
 
 def tween_request_profiling(handler, registry):
