@@ -1,4 +1,5 @@
 from __future__ import unicode_literals
+import getpass
 import posixpath
 import sys
 import py
@@ -10,6 +11,10 @@ from devpi_common.metadata import splitbasename, parse_version
 from devpi_common.url import URL
 from devpi_common.validation import validate_metadata, normalize_name
 from devpi_common.types import ensure_unicode, cached_property, parse_hash_spec
+try:
+    from itertools import zip_longest
+except ImportError:
+    from itertools import izip_longest as zip_longest
 from time import gmtime
 from .auth import hash_password, verify_and_update_password_hash
 from .filestore import FileEntry
@@ -17,6 +22,15 @@ from .log import threadlog, thread_current_log
 from .readonly import get_mutable_deepcopy
 
 
+def join_requires(links, requires_python):
+    # build list of (key, href, require_python) tuples
+    result = []
+    for link, require_python in zip_longest(links, requires_python, fillvalue=None):
+        key, href = link
+        result.append((key, href, require_python))
+    return result
+
+
 def run_passwd(root, username):
     user = root.get_user(username)
     log = thread_current_log()
@@ -24,8 +38,8 @@ def run_passwd(root, username):
         log.error("user %r not found" % username)
         return 1
     for i in range(3):
-        pwd = py.std.getpass.getpass("enter password for %s: " % user.name)
-        pwd2 = py.std.getpass.getpass("repeat password for %s: " % user.name)
+        pwd = getpass.getpass("enter password for %s: " % user.name)
+        pwd2 = getpass.getpass("repeat password for %s: " % user.name)
         if pwd != pwd2:
             log.error("password don't match")
         else:
@@ -82,6 +96,10 @@ class UpstreamError(ModelException):
     """ If an upstream could not be reached or didn't respond correctly. """
 
 
+class UpstreamNotFoundError(UpstreamError):
+    """ If upstream returned a not found error. """
+
+
 class MissesRegistration(ModelException):
     """ A prior registration of release metadata is required. """
 
@@ -214,14 +232,14 @@ def get_indexconfig(hooks, type, **kwargs):
     return ixconfig
 
 
-name_char_blacklist_regexp = re.compile(
-    '[\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f'
-    '\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'
-    ' !"#$%&\'()*+,/:;<=>?\[\\\\\]^`{|}~]')
+name_char_blocklist_regexp = re.compile(
+    r'[\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f'
+    r'\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'
+    r' !"#$%&\'()*+,/:;<=>?\[\\\\\]^`{|}~]')
 
 
 def is_valid_name(name):
-    return not name_char_blacklist_regexp.search(name)
+    return not name_char_blocklist_regexp.search(name)
 
 
 class User:
@@ -375,6 +393,7 @@ class BaseStage(object):
     InvalidUser = InvalidUser
     NotFound = NotFound
     UpstreamError = UpstreamError
+    UpstreamNotFoundError = UpstreamNotFoundError
     MissesRegistration = MissesRegistration
     MissesVersion = MissesVersion
     NonVolatile = NonVolatile
@@ -390,6 +409,23 @@ class BaseStage(object):
         self.keyfs = xom.keyfs
         self.filestore = xom.filestore
 
+    @cached_property
+    def user(self):
+        # only few methods need the user object.
+        return self.model.get_user(self.username)
+
+    def get(self):
+        userconfig = self.user.get()
+        return userconfig.get("indexes", {}).get(self.index)
+
+    def delete(self):
+        with self.user.key.update() as userconfig:
+            indexes = userconfig.get("indexes", {})
+            if self.index not in indexes:
+                threadlog.info("index %s not exists" % self.index)
+                return False
+            del indexes[self.index]
+
     def key_projsimplelinks(self, project):
         return self.keyfs.PROJSIMPLELINKS(user=self.username,
             index=self.index, project=normalize_name(project))
@@ -397,19 +433,22 @@ class BaseStage(object):
     def get_releaselinks(self, project):
         # compatibility access method used by devpi-web and tests
         project = normalize_name(project)
-        return [self._make_elink(project, key, href)
-                for key, href in self.get_simplelinks(project)]
+        try:
+            return [self._make_elink(project, key, href, require_python)
+                    for key, href, require_python in self.get_simplelinks(project)]
+        except self.UpstreamNotFoundError:
+            return []
 
     def get_releaselinks_perstage(self, project):
         # compatibility access method for devpi-findlinks and possibly other plugins
         project = normalize_name(project)
-        return [self._make_elink(project, key, href)
-                for key, href in self.get_simplelinks_perstage(project)]
+        return [self._make_elink(project, key, href, require_python)
+                for key, href, require_python in self.get_simplelinks_perstage(project)]
 
-    def _make_elink(self, project, key, href):
-        rp = SimplelinkMeta((key, href))
+    def _make_elink(self, project, key, href, require_python):
+        rp = SimplelinkMeta((key, href, require_python))
         linkdict = {"entrypath": rp._url.path, "hash_spec": rp._url.hash_spec,
-                    "eggfragment": rp.eggfragment}
+                "eggfragment": rp.eggfragment, "require_python": require_python}
         return ELink(self.filestore, linkdict, project, rp.version)
 
     def get_linkstore_perstage(self, name, version, readonly=True):
@@ -477,14 +516,19 @@ class BaseStage(object):
         """
         all_links = []
         seen = set()
-        for stage, res in self.op_sro_check_mirror_whitelist(
-            "get_simplelinks_perstage", project=project):
-            for key, href in res:
-                if key not in seen:
-                    seen.add(key)
-                    all_links.append((key, href))
+
+        try:
+            for stage, res in self.op_sro_check_mirror_whitelist(
+                    "get_simplelinks_perstage", project=project):
+                for key, href, require_python in res:
+                    if key not in seen:
+                        seen.add(key)
+                        all_links.append((key, href, require_python))
+        except self.UpstreamNotFoundError:
+            return []
+
         if sorted_links:
-           all_links = [(v.key, v.href)
+            all_links = [(v.key, v.href, v.require_python)
                         for v in sorted(map(SimplelinkMeta, all_links), reverse=True)]
         return all_links
 
@@ -500,7 +544,7 @@ class BaseStage(object):
                     has_mirror_base=has_mirror_base,
                     blocked_by_mirror_whitelist=stage.name if blocked_by_mirror_whitelist else None)
             private_hit = private_hit or in_index
-            whitelist = set(stage.ixconfig["mirror_whitelist"])
+            whitelist = set(stage.ixconfig.get("mirror_whitelist", set()))
             whitelisted = whitelisted or '*' in whitelist or project in whitelist
         return dict(
             has_mirror_base=False,
@@ -532,7 +576,7 @@ class BaseStage(object):
                     threadlog.debug("private package %r whitelisted at stage %s",
                                     project, whitelisted.name)
             else:
-                whitelist = set(stage.ixconfig["mirror_whitelist"])
+                whitelist = set(stage.ixconfig.get("mirror_whitelist", set()))
                 if '*' in whitelist or project in whitelist:
                     whitelisted = stage
                 elif stage.has_project_perstage(project):
@@ -542,7 +586,7 @@ class BaseStage(object):
                 res = getattr(stage, opname)(**kw)
                 private_hit = private_hit or res
                 yield stage, res
-            except UpstreamError as exc:
+            except self.UpstreamError as exc:
                 # If we are currently checking ourself raise the error, it is fatal
                 if stage is self:
                     raise
@@ -586,7 +630,9 @@ class PrivateStage(BaseStage):
         'provides', 'requires', 'obsoletes',
         # Metadata 1.2
         'project_urls', 'provides_dist', 'obsoletes_dist',
-        'requires_dist', 'requires_external', 'requires_python')
+        'requires_dist', 'requires_external', 'requires_python',
+        # Metadata 2.1
+        'description_content_type', 'provides_extras')
     metadata_list_fields = (
         'platform', 'classifiers', 'obsoletes',
         'requires', 'provides', 'obsoletes_dist',
@@ -598,11 +644,6 @@ class PrivateStage(BaseStage):
         super(PrivateStage, self).__init__(xom, username, index, ixconfig)
         self.key_projects = self.keyfs.PROJNAMES(user=username, index=index)
 
-    @cached_property
-    def user(self):
-        # only few methods need the user object.
-        return self.model.get_user(self.username)
-
     def modify(self, index=None, **kw):
         if 'type' in kw and self.ixconfig["type"] != kw['type']:
             raise InvalidIndexconfig(
@@ -621,21 +662,11 @@ class PrivateStage(BaseStage):
             self.ixconfig = newconfig
             return newconfig
 
-    def get(self):
-        userconfig = self.user.get()
-        return userconfig.get("indexes", {}).get(self.index)
-
     def delete(self):
         # delete all projects on this index
         for name in self.list_projects_perstage():
             self.del_project(name)
-        with self.user.key.update() as userconfig:
-            indexes = userconfig.get("indexes", {})
-            if self.index not in indexes:
-                threadlog.info("index %s not exists" % self.index)
-                return False
-            del indexes[self.index]
-
+        BaseStage.delete(self)
 
     #
     # registering project and version metadata
@@ -678,6 +709,10 @@ class PrivateStage(BaseStage):
         if version not in versions:
             versions.add(version)
             self.key_projversions(project).set(versions)
+        self.add_project_name(project)
+
+    def add_project_name(self, project):
+        project = normalize_name(project)
         projects = self.key_projects.get(readonly=False)
         if project not in projects:
             projects.add(project)
@@ -712,6 +747,19 @@ class PrivateStage(BaseStage):
                 self.del_project(project)
             self._regen_simplelinks(project)
 
+    def del_entry(self, entry, cleanup=True):
+        # we need to store project and version for use in cleanup part below
+        project = entry.project
+        version = entry.version
+        linkstore = self.get_linkstore_perstage(
+            project, version, readonly=False)
+        linkstore.remove_links(basename=entry.basename)
+        entry.delete()
+        if cleanup:
+            if not linkstore.get_links():
+                self.del_versiondata(project, version)
+            self._regen_simplelinks(project)
+
     def list_versions_perstage(self, project):
         return self.key_projversions(project).get()
 
@@ -720,15 +768,24 @@ class PrivateStage(BaseStage):
         return self.key_projversion(project, version).get(readonly=readonly)
 
     def get_simplelinks_perstage(self, project):
-        return self.key_projsimplelinks(project).get().get("links", [])
+        data = self.key_projsimplelinks(project).get()
+        links = data.get("links", [])
+        requires_python = data.get("requires_python", [])
+        return join_requires(links, requires_python)
 
     def _regen_simplelinks(self, project_input):
         project = normalize_name(project_input)
         links = []
+        requires_python = []
         for version in self.list_versions_perstage(project):
             linkstore = self.get_linkstore_perstage(project, version)
-            links.extend(map(make_key_and_href, linkstore.get_links("releasefile")))
-        self.key_projsimplelinks(project).set({"links": links})
+            releases = linkstore.get_links("releasefile")
+            links.extend(map(make_key_and_href, releases))
+            require_python = self.get_versiondata_perstage(project,
+                    version).get('requires_python')
+            requires_python.extend([require_python] * len(releases))
+        data_dict = {u"links":links, u"requires_python":requires_python}
+        self.key_projsimplelinks(project).set(data_dict)
 
     def list_projects_perstage(self):
         return self.key_projects.get()
@@ -874,6 +931,14 @@ class LinkStore:
             raise MissesRegistration("%s-%s on stage %s",
                                      project, version, stage.name)
 
+    @property
+    def metadata(self):
+        metadata = {}
+        for k, v in get_mutable_deepcopy(self.verdata).items():
+            if not k.startswith("+"):
+                metadata[k] = v
+        return metadata
+
     def get_file_entry(self, relpath):
         return self.filestore.get_file_entry(relpath)
 
@@ -971,7 +1036,7 @@ class LinkStore:
 class SimplelinkMeta(CompareMixin):
     """ helper class to provide information for items from get_simplelinks() """
     def __init__(self, key_href):
-        self.key, self.href = key_href
+        self.key, self.href, self.require_python = key_href
         self._url = URL(self.href)
         self.name, self.version, self.ext = splitbasename(self._url.basename, checkarch=False)
         self.eggfragment = self._url.eggfragment
@@ -1091,6 +1156,10 @@ class EventSubscribers:
             entry = FileEntry(self.xom, ev.typedkey, meta=ev.value)
             if not entry.project or not entry.version:
                 # the entry was deleted
+                self.xom.config.hook.devpiserver_on_remove_file(
+                    stage=stage,
+                    relpath=ev.typedkey.relpath
+                )
                 return
             name = entry.project
             assert name == normalize_name(name)
