@@ -402,7 +402,7 @@ class ImportFileReplica:
     def __call__(self, fswriter, key, val, back_serial):
         threadlog.debug("ImportFileReplica for %s, %s", key, val)
         relpath = key.relpath
-        entry = self.xom.filestore.get_file_entry_raw(key, val)
+        entry = self.xom.filestore.get_file_entry_from_key(key, meta=val)
         file_exists = fswriter.conn.io_file_exists(entry._storepath)
         if val is None:
             if back_serial >= 0:
@@ -458,9 +458,10 @@ class FileReplicationError(Exception):
     def __init__(self, response, relpath, message=None):
         self.url = response.url
         self.status_code = response.status_code
+        self.reason = response.reason
         self.relpath = relpath
         self.message = message or "failed"
 
     def __str__(self):
-        return "FileReplicationError with %s, code=%s, relpath=%s, message=%s" % (
-               self.url, self.status_code, self.relpath, self.message)
+        return "FileReplicationError with %s, code=%s, reason=%s, relpath=%s, message=%s" % (
+               self.url, self.status_code, self.reason, self.relpath, self.message)
