@@ -9,6 +9,8 @@ from devpi_common.types import ensure_unicode
 from devpi_common.url import URL
 from devpi_common.metadata import get_pyversion_filetype
 import devpi_server
+from pluggy import HookimplMarker
+from pyramid.compat import escape
 from pyramid.compat import urlparse
 from pyramid.interfaces import IAuthenticationPolicy
 from pyramid.httpexceptions import HTTPException, HTTPFound, HTTPSuccessful
@@ -31,6 +33,7 @@ from .log import thread_push_log, thread_pop_log, threadlog
 
 from .auth import Auth
 
+devpiweb_hookimpl = HookimplMarker("devpiweb")
 server_version = devpi_server.__version__
 
 
@@ -160,26 +163,28 @@ def tween_request_logging(handler, registry):
     def request_log_handler(request):
         tag = "[req%s]" %(next(req_count))
         log = thread_push_log(tag)
-        request.log = log
-        log.info("%s %s" % (request.method, request.path,))
-        now = time()
-        response = handler(request)
-        duration = time() - now
-        rheaders = response.headers
-        serial = rheaders.get("X-DEVPI-SERIAL")
-        rheaders.update(meta_headers)
-        uuid, master_uuid = make_uuid_headers(nodeinfo)
-        rheaders[str("X-DEVPI-UUID")] = str(uuid)
-        rheaders[H_MASTER_UUID] = str(master_uuid)
+        try:
+            request.log = log
+            log.info("%s %s" % (request.method, request.path,))
+            now = time()
+            response = handler(request)
+            duration = time() - now
+            rheaders = response.headers
+            serial = rheaders.get("X-DEVPI-SERIAL")
+            rheaders.update(meta_headers)
+            uuid, master_uuid = make_uuid_headers(nodeinfo)
+            rheaders[str("X-DEVPI-UUID")] = str(uuid)
+            rheaders[H_MASTER_UUID] = str(master_uuid)
 
-        log.debug("%s %.3fs serial=%s length=%s type=%s",
-                  response.status_code,
-                  duration,
-                  serial,
-                  rheaders.get("content-length"),
-                  rheaders.get("content-type"),
-        )
-        thread_pop_log(tag)
+            log.debug("%s %.3fs serial=%s length=%s type=%s",
+                      response.status,
+                      duration,
+                      serial,
+                      rheaders.get("content-length"),
+                      rheaders.get("content-type"),
+            )
+        finally:
+            thread_pop_log(tag)
         return response
     return request_log_handler
 
@@ -217,6 +222,23 @@ def set_header_devpi_serial(response, tx):
 def is_mutating_http_method(method):
     return method in ("PUT", "POST", "PATCH", "DELETE", "PUSH")
 
+
+def get_actions(json):
+    result = []
+    for item in json:
+        key, sep, value = item.partition('=')
+        if key.endswith('+'):
+            op = 'add'
+            key = key[:-1]
+        elif key.endswith('-'):
+            op = 'del'
+            key = key[:-1]
+        else:
+            op = 'set'
+        result.append((op, key, value))
+    return result
+
+
 class StatusView:
     def __init__(self, request):
         self.request = request
@@ -266,6 +288,7 @@ class StatusView:
         apireturn(200, type="status", result=self._status())
 
 
+@devpiweb_hookimpl
 def devpiweb_get_status_info(request):
     msgs = []
     status = StatusView(request)._status()
@@ -354,6 +377,7 @@ class PyPIView:
         api = {
             "login": request.route_url('/+login'),
             "authstatus": self.get_auth_status(),
+            "features": self.xom.supported_features,
         }
         if path:
             parts = path.split("/")
@@ -393,7 +417,30 @@ class PyPIView:
     # index serving and upload
     #
 
+    @property
+    def _use_absolute_urls(self):
+        if self.xom.config.args.absolute_urls:
+            return True
+        if 'HTTP_X_DEVPI_ABSOLUTE_URLS' in self.request.environ:
+            return True
+        return False
+
     @view_config(route_name="/{user}/{index}/+simple/{project}")
+    def simple_list_project_redirect(self):
+        """
+        PEP 503:
+        the repository SHOULD redirect the URLs without a /
+        to add a / to the end
+        """
+        request = self.request
+        abort_if_invalid_project(request, request.matchdict["project"])
+        return HTTPFound(location=self.request.route_url(
+            "/{user}/{index}/+simple/{project}/",
+            user=self.context.username,
+            index=self.context.index,
+            project=self.context.project))
+
+    @view_config(route_name="/{user}/{index}/+simple/{project}/")
     def simple_list_project(self):
         request = self.request
         abort_if_invalid_project(request, request.matchdict["project"])
@@ -433,7 +480,7 @@ class PyPIView:
         response.content_type = "text/html ; charset=utf-8"
 
         title = "%s: links for %s" % (stage.name, project)
-        yield ("<html><head><title>%s</title></head><body><h1>%s</h1>\n" %
+        yield ("<!DOCTYPE html><html><head><title>%s</title></head><body><h1>%s</h1>\n" %
                (title, title)).encode("utf-8")
 
         if embed_form:
@@ -445,12 +492,38 @@ class PyPIView:
                    "<strong>%s</strong> are included.</p>"
                    % blocked_index).encode('utf-8')
 
-        url = URL(self.request.path_info)
-        for key, href in result:
-            yield ('%s <a href="%s">%s</a><br/>\n' %
-                   ("/".join(href.split("/", 2)[:2]),
-                    url.relpath("/" + href),
-                    key)).encode("utf-8")
+        if self._use_absolute_urls:
+            # for joinpath we need the root url
+            application_url = self.request.application_url
+            if not application_url.endswith("/"):
+                application_url = application_url + "/"
+            url = URL(application_url)
+
+            def make_url(href):
+                return url.joinpath(href).url
+        else:
+            # for relpath we need the path of the current page
+            url = URL(self.request.path_info)
+
+            def make_url(href):
+                return url.relpath("/" + href)
+
+        for key, href, require_python in result:
+            stage = "/".join(href.split("/", 2)[:2])
+            relurl = make_url(href)
+            if require_python is None:
+                data = dict(stage=stage, url=relurl, key=key)
+                yield ('{stage} <a href="{url}">{key}</a><br/>\n'
+                    .format(**data).encode('utf-8')
+                    )
+            else:
+                require_python = escape(require_python)
+                data = dict(stage=stage, url=relurl, key=key,
+                    require_python=require_python)
+                yield ('{stage} <a data-requires-python="{require_python}" '
+                    'href="{url}">{key}</a><br/>\n'
+                    .format(**data).encode('utf-8')
+                    )
 
         yield "</body></html>".encode("utf-8")
 
@@ -463,6 +536,18 @@ class PyPIView:
         submit = '<input name="refresh" type="submit" value="%s"/>' % title
         return '<form action="%s" method="post">%s</form>' % (url, submit)
 
+    @view_config(route_name="/{user}/{index}/+simple")
+    def simple_list_all_redirect(self):
+        """
+        PEP 503:
+        the repository SHOULD redirect the URLs without a /
+        to add a / to the end
+        """
+        return HTTPFound(location=self.request.route_url(
+            "/{user}/{index}/+simple/",
+            user=self.context.username,
+            index=self.context.index))
+
     @view_config(route_name="/{user}/{index}/+simple/")
     def simple_list_all(self):
         self.log.info("starting +simple")
@@ -481,7 +566,7 @@ class PyPIView:
         response.content_type = "text/html ; charset=utf-8"
         title =  "%s: simple list (including inherited indices)" %(
                  stage.name)
-        yield ("<html><head><title>%s</title></head><body><h1>%s</h1>" %(
+        yield ("<!DOCTYPE html><html><head><title>%s</title></head><body><h1>%s</h1>" %(
               title, title)).encode("utf-8")
         all_names = set()
         for stage, names in stage_results:
@@ -492,10 +577,10 @@ class PyPIView:
             yield ("<h2>" + h2 + "</h2>").encode("utf-8")
             for name in sorted(names):
                 if name not in all_names:
-                    anchor = '<a href="%s">%s</a><br/>\n' % (name, name)
+                    anchor = '<a href="%s/">%s</a><br/>\n' % (name, name)
                     yield anchor.encode("utf-8")
                     all_names.add(name)
-        yield "</body>".encode("utf-8")
+        yield "</body></html>".encode("utf-8")
 
     @view_config(
         route_name="/{user}/{index}/+simple/{project}/refresh", request_method="POST")
@@ -507,7 +592,7 @@ class PyPIView:
             stage.clear_simplelinks_cache(context.project)
             stage.get_simplelinks_perstage(context.project)
         return HTTPFound(location=self.request.route_url(
-            "/{user}/{index}/+simple/{project}",
+            "/{user}/{index}/+simple/{project}/",
             user=context.username, index=context.index, project=context.project))
 
     @view_config(
@@ -550,7 +635,28 @@ class PyPIView:
         permission="index_modify")
     def index_modify(self):
         stage = self.context.stage
-        kvdict = getkvdict_index(self.xom.config.hook, getjson(self.request))
+        json = _getjson(self.request)
+        if isinstance(json, list):
+            ixconfig = stage.get()
+            for op, key, value in get_actions(json):
+                if op == 'del':
+                    if value not in ixconfig[key]:
+                        apireturn(
+                            400, "The '%s' setting doesn't have value '%s'" % (key, value))
+                    if isinstance(ixconfig[key], tuple):
+                        ixconfig[key] = tuple(
+                            x for x in ixconfig[key] if x != value)
+                    else:
+                        ixconfig[key].remove(value)
+                elif op == 'add':
+                    if isinstance(ixconfig[key], tuple):
+                        ixconfig[key] += (value,)
+                    else:
+                        ixconfig[key].append(value)
+                elif op == 'set':
+                    ixconfig[key] = value
+            json = ixconfig
+        kvdict = getkvdict_index(self.xom.config.hook, json)
         try:
             ixconfig = stage.modify(**kvdict)
         except InvalidIndexconfig as e:
@@ -571,6 +677,9 @@ class PyPIView:
     @view_config(route_name="/{user}/{index}", request_method=("POST", "PUSH"))
     def pushrelease(self):
         request = self.request
+        if request.POST.get(':action'):
+            # this is actually a submit
+            return self.submit()
         stage = self.context.stage
         pushdata = getjson(request)
         try:
@@ -593,7 +702,7 @@ class PyPIView:
             apireturn(404, message="no release/files found for %s-%s" %(
                       name, version))
 
-        metadata = get_pure_metadata(linkstore.verdata)
+        metadata = linkstore.metadata
 
         results = []
         targetindex = pushdata.get("targetindex", None)
@@ -856,8 +965,13 @@ class PyPIView:
 
     @view_config(route_name="simple_redirect")
     def simple_redirect(self):
+        """
+        PEP 503:
+        the repository SHOULD redirect the URLs without a /
+        to add a / to the end
+        """
         return HTTPFound(location=self.request.route_url(
-            "/{user}/{index}/+simple/{project}",
+            "/{user}/{index}/+simple/{project}/",
             user=self.context.username,
             index=self.context.index,
             project=self.context.project))
@@ -880,8 +994,6 @@ class PyPIView:
         permission="del_project")
     def del_project(self):
         stage = self.context.stage
-        if stage.ixconfig["type"] == "mirror":
-            abort(self.request, 405, "cannot delete on mirror index")
         project = self.context.project
         if not stage.ixconfig["volatile"]:
             apireturn(403, "project %r is on non-volatile index %s" %(
@@ -926,8 +1038,6 @@ class PyPIView:
     def del_versiondata(self):
         stage = self.context.stage
         name, version = self.context.project, self.context.version
-        if stage.ixconfig["type"] == "mirror":
-            abort(self.request, 405, "cannot delete on mirror index")
         if not stage.ixconfig["volatile"]:
             abort(self.request, 403, "cannot delete version on non-volatile index")
         try:
@@ -936,25 +1046,22 @@ class PyPIView:
             abort(self.request, 404, e.msg)
         apireturn(200, "project %r version %r deleted" % (name, version))
 
-    @view_config(route_name="/{user}/{index}/+e/{relpath:.*}")
-    @view_config(route_name="/{user}/{index}/+f/{relpath:.*}")
-    def pkgserv(self):
-        request = self.request
-        relpath = request.path_info.strip("/")
+    def _relpath_from_request(self):
+        relpath = self.request.path_info.strip("/")
         if "#" in relpath:   # XXX unclear how this happens (did with bottle)
             relpath = relpath.split("#", 1)[0]
-        filestore = self.xom.filestore
-        entry = filestore.get_file_entry(relpath)
+        return relpath
+
+    def _pkgserv(self, entry):
+        request = self.request
+        if entry is None:
+            abort(request, 404, "no such file")
+        elif not entry.meta:
+            abort(request, 410, "file existed, deleted in later serial")
+
         if json_preferred(request):
-            if not entry or not entry.meta:
-                apireturn(404, "no such release file")
             entry_data = get_mutable_deepcopy(entry.meta)
             apireturn(200, type="releasefilemeta", result=entry_data)
-        if not entry or not entry.meta:
-            if entry is None:
-                abort(request, 404, "no such file")
-            else:
-                abort(request, 410, "file existed, deleted in later serial")
 
         try:
             if should_fetch_remote_file(entry, request.headers):
@@ -973,6 +1080,47 @@ class PyPIView:
             content = entry.file_get_content()
             return Response(body=content, headers=headers)
 
+    @view_config(route_name="/{user}/{index}/+e/{relpath:.*}")
+    def mirror_pkgserv(self):
+        relpath = self._relpath_from_request()
+        # when a release is deleted from a mirror, we update the metadata,
+        # hence the key won't exist anymore, but we don't delete the file.
+        # We want people to notice that condition by returning a 404, but
+        # they can still recover the deleted release from the filesystem
+        # manually in case they need it.
+        key = self.xom.filestore.get_key_from_relpath(relpath)
+        if not key.exists():
+            abort(self.request, 404, "no such file")
+        entry = self.xom.filestore.get_file_entry_from_key(key)
+        return self._pkgserv(entry)
+
+    @view_config(route_name="/{user}/{index}/+f/{relpath:.*}")
+    def stage_pkgserv(self):
+        relpath = self._relpath_from_request()
+        entry = self.xom.filestore.get_file_entry(relpath)
+        return self._pkgserv(entry)
+
+    @view_config(route_name="/{user}/{index}/+e/{relpath:.*}",
+                 permission="del_entry",
+                 request_method="DELETE")
+    @view_config(route_name="/{user}/{index}/+f/{relpath:.*}",
+                 permission="del_entry",
+                 request_method="DELETE")
+    def del_pkg(self):
+        stage = self.context.stage
+        if not stage.ixconfig["volatile"]:
+            abort(self.request, 403, "cannot delete version on non-volatile index")
+        relpath = self.request.path_info.strip("/")
+        filestore = self.xom.filestore
+        entry = filestore.get_file_entry(relpath)
+        if entry is None:
+            abort(self.request, 404, "package %r doesn't exist" % relpath)
+        try:
+            stage.del_entry(entry)
+        except stage.NotFound as e:
+            abort(self.request, 404, e.msg)
+        apireturn(200, "package %r deleted" % relpath)
+
     @view_config(route_name="/{user}/{index}", accept="application/json", request_method="GET")
     def index_get(self):
         stage = self.context.stage
@@ -1108,16 +1256,25 @@ def url_for_entrypath(request, entrypath):
         route_name, user=user, index=index, relpath=relpath)
 
 
-def getjson(request, allowed_keys=None):
+def _getjson(request):
     try:
-        dict = request.json_body
+        d = request.json_body
     except ValueError:
         abort(request, 400, "Bad request: could not decode json")
+    return d
+
+
+def _check_allowed_keys(request, d, allowed_keys):
+    diff = set(d).difference(allowed_keys)
+    if diff:
+        abort(request, 400, "json keys not recognized: %s" % ",".join(diff))
+
+
+def getjson(request, allowed_keys=None):
+    d = _getjson(request)
     if allowed_keys is not None:
-        diff = set(dict).difference(allowed_keys)
-        if diff:
-            abort(request, 400, "json keys not recognized: %s" % ",".join(diff))
-    return dict
+        _check_allowed_keys(request, d, allowed_keys)
+    return d
 
 
 def abort_if_invalid_filename(request, name, filename):
@@ -1146,10 +1303,3 @@ def getkvdict_index(hook, req):
             kvdict[key] = req[key]
     return kvdict
 
-def get_pure_metadata(somedict):
-    metadata = {}
-    for n, v in somedict.items():
-        if n[0] != "+":
-            metadata[n] = v
-    return metadata
-
