@@ -1,6 +1,7 @@
 from __future__ import print_function
 import re
 from webtest.forms import Upload
+import json
 import webtest
 import mimetypes
 import subprocess
@@ -10,6 +11,7 @@ import py
 import requests
 import socket
 import sys
+import time
 from bs4 import BeautifulSoup
 from contextlib import closing
 from devpi_server import extpypi
@@ -19,7 +21,8 @@ from devpi_common.url import URL
 from devpi_server.extpypi import PyPIStage
 from devpi_server.log import threadlog, thread_clear_log
 from pyramid.authentication import b64encode
-from time import sleep
+from pyramid.compat import escape
+from pyramid.httpexceptions import status_map
 
 import hashlib
 try:
@@ -70,6 +73,7 @@ def Queue():
 
 @pytest.fixture()
 def caplog(caplog):
+    import logging
     """ enrich the pytest-catchlog funcarg. """
     def getrecords(msgrex=None, minlevel="DEBUG"):
         if msgrex is not None:
@@ -85,6 +89,7 @@ def caplog(caplog):
             recs.append(rec)
         return recs
     caplog.getrecords = getrecords
+    caplog.set_level(logging.NOTSET)
     return caplog
 
 @pytest.fixture
@@ -93,7 +98,7 @@ def gentmp(request):
     cache = []
     def gentmp(name=None):
         if not cache:
-            prefix = re.sub("[\W]", "_", request.node.name)
+            prefix = re.sub(r"[\W]", "_", request.node.name)
             basedir = tmpdirhandler.mktemp(prefix, numbered=True)
             cache.append(basedir)
         else:
@@ -109,12 +114,12 @@ def gentmp(request):
 def auto_transact(request):
     names = request.fixturenames
     if ("xom" not in names and "keyfs" not in names) or (
-        request.node.get_marker("notransaction")):
+        request.node.get_closest_marker("notransaction")):
         yield
         return
     keyfs = request.getfixturevalue("keyfs")
 
-    write = True if request.node.get_marker("writetransaction") else False
+    write = True if request.node.get_closest_marker("writetransaction") else False
     keyfs.begin_transaction_in_thread(write=write)
     yield
     try:
@@ -200,27 +205,27 @@ def makexom(request, gentmp, httpget, monkeypatch, storage_info):
             pm.register(plugin)
         serverdir = gentmp()
         fullopts = ["devpi-server", "--serverdir", serverdir] + list(opts)
-        if request.node.get_marker("with_replica_thread"):
+        if request.node.get_closest_marker("with_replica_thread"):
             fullopts.append("--master=http://localhost")
-        if not request.node.get_marker("no_storage_option"):
+        if not request.node.get_closest_marker("no_storage_option"):
             if storage_info["name"] != "sqlite":
                 fullopts.append("--storage=%s" % storage_info["name"])
         fullopts = [str(x) for x in fullopts]
         config = parseoptions(pm, fullopts)
         config.init_nodeinfo()
         for marker in ("storage_with_filesystem",):
-            if request.node.get_marker(marker):
+            if request.node.get_closest_marker(marker):
                 info = config._storage_info()
                 markers = info.get("_test_markers", [])
                 if marker not in markers:
                     pytest.skip("The storage doesn't have marker '%s'." % marker)
-        if not request.node.get_marker("no_storage_option"):
+        if not request.node.get_closest_marker("no_storage_option"):
             assert storage_info["storage"] is config.storage
-        if request.node.get_marker("nomocking"):
+        if request.node.get_closest_marker("nomocking"):
             xom = XOM(config)
         else:
             xom = XOM(config, httpget=httpget)
-            if not request.node.get_marker("nomockprojectsremote"):
+            if not request.node.get_closest_marker("nomockprojectsremote"):
                 monkeypatch.setattr(extpypi.PyPIStage, "_get_remote_projects",
                     lambda self: set())
             add_pypistage_mocks(monkeypatch, httpget)
@@ -229,15 +234,15 @@ def makexom(request, gentmp, httpget, monkeypatch, storage_info):
         if not xom.config.args.master_url:
             with xom.keyfs.transaction(write=True):
                 set_default_indexes(xom.model)
-        if request.node.get_marker("with_replica_thread"):
+        if request.node.get_closest_marker("with_replica_thread"):
             from devpi_server.replica import ReplicaThread
             rt = ReplicaThread(xom)
             xom.replica_thread = rt
             xom.thread_pool.register(rt)
             xom.thread_pool.start_one(rt)
-        if request.node.get_marker("start_threads"):
+        if request.node.get_closest_marker("start_threads"):
             xom.thread_pool.start()
-        elif request.node.get_marker("with_notifier"):
+        elif request.node.get_closest_marker("with_notifier"):
             xom.thread_pool.start_one(xom.keyfs.notifier)
         request.addfinalizer(xom.thread_pool.shutdown)
         return xom
@@ -283,9 +288,13 @@ def makemapp(request, maketestapp, makexom):
 
 
 def make_simple_pkg_info(name, text="", pkgver=None, hash_type=None,
-                         pypiserial=None):
+                         pypiserial=None, requires_python=None):
     class ret:
         hash_spec = ""
+    if requires_python:
+        requires_python = ' data-requires-python="%s"' % escape(requires_python)
+    else:
+        requires_python = ''
     if pkgver is not None:
         assert not text
         if hash_type and "#" not in pkgver:
@@ -293,7 +302,8 @@ def make_simple_pkg_info(name, text="", pkgver=None, hash_type=None,
             hash_value = getattr(hashlib, hash_type)(hv).hexdigest()
             ret.hash_spec = "%s=%s" %(hash_type, hash_value)
             pkgver += "#" + ret.hash_spec
-        text = '<a href="../../{name}/{pkgver}">{pkgver}</a>'.format(name=name, pkgver=pkgver)
+        text = '<a href="../../{name}/{pkgver}"{requires_python}>{pkgver}</a>'.format(
+            name=name, pkgver=pkgver, requires_python=requires_python)
     elif text and "{md5}" in text:
         text = text.format(md5=getmd5(text))
     elif text and "{sha256}" in text:
@@ -306,7 +316,7 @@ def httpget(pypiurls):
     class MockHTTPGet:
         def __init__(self):
             self.url2response = {}
-            self._md5 = py.std.hashlib.md5()
+            self._md5 = hashlib.md5()
 
         def __call__(self, url, allow_redirects=False, extra_headers=None):
             class mockresponse:
@@ -321,10 +331,6 @@ def httpget(pypiurls):
                     if "content" in fakeresponse:
                         xself.raw = py.io.BytesIO(fakeresponse["content"])
 
-                @property
-                def status(xself):
-                    return "%s" % xself.status_code
-
                 def __repr__(xself):
                     return "<mockresponse %s url=%s>" % (xself.status_code,
                                                          xself.url)
@@ -333,16 +339,20 @@ def httpget(pypiurls):
             return r
 
         def mockresponse(self, mockurl, **kw):
-            if "status_code" not in kw:
-                kw["status_code"] = 200
+            kw.setdefault("status_code", 200)
+            kw.setdefault("reason", getattr(
+                status_map.get(kw["status_code"]),
+                "title",
+                "Devpi Mock Error"))
             log.debug("set mocking response %s %s", mockurl, kw)
             self.url2response[mockurl] = kw
 
         def mock_simple(self, name, text="", pkgver=None, hash_type=None,
-                        pypiserial=10000, remoteurl=None, **kw):
+                        pypiserial=10000, remoteurl=None, requires_python=None,
+                        **kw):
             ret, text = make_simple_pkg_info(
                 name, text=text, pkgver=pkgver, hash_type=hash_type,
-                pypiserial=pypiserial)
+                pypiserial=pypiserial, requires_python=requires_python)
             if remoteurl is None:
                 remoteurl = pypiurls.simple
             headers = kw.setdefault("headers", {})
@@ -369,21 +379,17 @@ def keyfs(xom):
 def model(xom):
     return xom.model
 
-@pytest.fixture
-def pypistage(xom):
-    from devpi_server.main import _pypi_ixconfig_default
-    return PyPIStage(xom, username="root", index="pypi",
-                     ixconfig=_pypi_ixconfig_default)
 
 @pytest.fixture
-def replica_pypistage(replica_xom):
-    return pypistage(replica_xom)
+def pypistage(devpiserver_makepypistage, xom):
+    return devpiserver_makepypistage(xom)
+
 
 def add_pypistage_mocks(monkeypatch, httpget):
     # add some mocking helpers
     PyPIStage.url2response = httpget.url2response
     def mock_simple(self, name, text=None, pypiserial=10000, **kw):
-        self.cache_link_updates.expire(name)
+        self.cache_retrieve_times.expire(name)
         return self.httpget.mock_simple(name,
                  text=text, pypiserial=pypiserial, **kw)
     monkeypatch.setattr(PyPIStage, "mock_simple", mock_simple, raising=False)
@@ -401,7 +407,7 @@ def add_pypistage_mocks(monkeypatch, httpget):
                    "content-type": mimetypes.guess_type(path),
                    "last-modified": "today",}
         url = URL(self.mirror_url).joinpath(path)
-        return self.httpget.mockresponse(url.url, raw=py.io.BytesIO(content),
+        return self.httpget.mockresponse(url.url, content=content,
                                          headers=headers, **kw)
     monkeypatch.setattr(PyPIStage, "mock_extfile", mock_extfile, raising=False)
 
@@ -711,7 +717,7 @@ class Mapp(MappMixin):
         indexname = self._getindexname(indexname)
         req = dict(name=name, version=version, targetindex=index)
         r = self.testapp.push(
-            '/%s' % indexname, py.std.json.dumps(req), expect_errors=True)
+            '/%s' % indexname, json.dumps(req), expect_errors=True)
         assert r.status_code == code
         return r
 
@@ -743,7 +749,7 @@ class Mapp(MappMixin):
         return r
 
     def get_simple(self, project, code=200):
-        r = self.testapp.get(self.api.simpleindex + project,
+        r = self.testapp.get(self.api.simpleindex + project + '/',
                              expect_errors=True)
         assert r.status_code == code
         return r
@@ -978,7 +984,7 @@ def wait_for_port(host, port, timeout=60):
             s.settimeout(1)
             if s.connect_ex((host, port)) == 0:
                 return
-        sleep(1)
+        time.sleep(1)
         timeout -= 1
     raise RuntimeError(
         "The port %s on host %s didn't become accessible" % (port, host))
@@ -1003,11 +1009,11 @@ def call_devpi_in_dir():
         from _pytest.monkeypatch import MonkeyPatch
         from _pytest.pytester import RunResult
         m = MonkeyPatch()
-        m.setenv("DEVPI_SERVERDIR", server_dir)
+        m.setenv("DEVPISERVER_SERVERDIR", server_dir)
         m.setattr("sys.argv", [devpiserver])
         cap = py.io.StdCaptureFD()
         cap.startall()
-        now = py.std.time.time()
+        now = time.time()
         try:
             main(args)
         finally:
@@ -1015,7 +1021,7 @@ def call_devpi_in_dir():
             out, err = cap.reset()
             del cap
         return RunResult(
-            0, out.split("\n"), err.split("\n"), py.std.time.time() - now)
+            0, out.split("\n"), err.split("\n"), time.time() - now)
 
     return devpi
 
@@ -1207,11 +1213,11 @@ class SimPyPI:
         self.projects.pop(name)
 
     def add_release(self, name, title=None, text="", pkgver=None, hash_type=None,
-                    pypiserial=None, **kw):
+                    pypiserial=None, requires_python=None, **kw):
         project = self.add_project(name, title=title)
         ret, text = make_simple_pkg_info(
             name, text=text, pkgver=pkgver, hash_type=hash_type,
-            pypiserial=pypiserial)
+            pypiserial=pypiserial, requires_python=requires_python)
         assert text
         project['releases'].add(text.encode('utf-8'))
 
@@ -1337,10 +1343,10 @@ def gen():
 
 class Gen:
     def __init__(self):
-        self._md5 = py.std.hashlib.md5()
+        self._md5 = hashlib.md5()
 
     def pypi_package_link(self, pkgname, md5=True):
-        link = "https://pypi.python.org/package/some/%s" % pkgname
+        link = "https://pypi.org/package/some/%s" % pkgname
         if md5 == True:
             self._md5.update(link.encode("utf8"))  # basically random
             link += "#md5=%s" % self._md5.hexdigest()
