@@ -1,6 +1,10 @@
-import py
+import itsdangerous
 import pytest
-from devpi_server.auth import *
+from devpi_server.auth import getpwhash
+from devpi_server.auth import hash_password
+from devpi_server.auth import newsalt
+from devpi_server.auth import verify_and_update_password_hash
+from devpi_server.config import hookimpl
 
 pytestmark = [pytest.mark.writetransaction]
 
@@ -28,7 +32,7 @@ class TestAuth:
         model.create_user(username, password=password)
         proxy = auth.new_proxy_auth(username, password)
 
-        def r(*args, **kw): raise py.std.itsdangerous.SignatureExpired("123")
+        def r(*args, **kw): raise itsdangerous.SignatureExpired("123")
         monkeypatch.setattr(auth.serializer, "loads", r)
 
         assert auth._get_auth_status(username, proxy["password"]) == dict(status="expired")
@@ -70,6 +74,7 @@ class TestAuthPlugin:
     @pytest.fixture
     def plugin(self):
         class Plugin:
+            @hookimpl
             def devpiserver_auth_user(self, userdict, username, password):
                 return self.results.pop()
         return Plugin()
@@ -133,6 +138,7 @@ class TestAuthPlugins:
     @pytest.fixture
     def plugin1(self):
         class Plugin:
+            @hookimpl
             def devpiserver_auth_user(self, userdict, username, password):
                 return self.results.pop()
         return Plugin()
@@ -140,6 +146,7 @@ class TestAuthPlugins:
     @pytest.fixture
     def plugin2(self):
         class Plugin:
+            @hookimpl
             def devpiserver_auth_user(self, userdict, username, password):
                 return self.results.pop()
         return Plugin()
