@@ -1,4 +1,5 @@
 from __future__ import unicode_literals
+import requests.exceptions
 import time
 import hashlib
 import pytest
@@ -8,8 +9,13 @@ from devpi_server.extpypi import ProjectNamesCache, ProjectUpdateCache
 from test_devpi_server.conftest import getmd5
 
 
+def getlinks(text):
+    from bs4 import BeautifulSoup
+    return BeautifulSoup(text, "html.parser").findAll("a")
+
+
 class TestIndexParsing:
-    simplepy = URL("http://pypi.python.org/simple/py/")
+    simplepy = URL("https://pypi.org/simple/py/")
 
     @pytest.mark.parametrize("hash_type,hash_value", [
         ("sha256", "090123"),
@@ -37,7 +43,7 @@ class TestIndexParsing:
         assert link.url.endswith("/~user/py-1.4.12.zip#md5=12ab")
 
     def test_parse_index_simple_nocase(self):
-        simplepy = URL("http://pypi.python.org/simple/Py/")
+        simplepy = URL("https://pypi.org/simple/Py/")
         result = parse_index(simplepy,
             """<a href="../../pkg/py-1.4.12.zip#md5=12ab">qwe</a>
                <a href="../../pkg/PY-1.4.13.zip">qwe</a>
@@ -46,7 +52,7 @@ class TestIndexParsing:
         assert len(result.releaselinks) == 3
 
     def test_parse_index_simple_dir_egg_issue63(self):
-        simplepy = URL("http://pypi.python.org/simple/py/")
+        simplepy = URL("https://pypi.org/simple/py/")
         result = parse_index(simplepy,
             """<a href="../../pkg/py-1.4.12.zip#md5=12ab">qwe</a>
                <a href="../../pkg/#egg=py-dev">qwe</a>
@@ -66,7 +72,7 @@ class TestIndexParsing:
             import urlparse
         monkeypatch.setattr(urlparse, "uses_fragment",
                             urlparse.uses_fragment + ["svn"])
-        simplepy = URL("https://pypi.python.org/simple/zope.sqlalchemy/")
+        simplepy = URL("https://pypi.org/simple/zope.sqlalchemy/")
         result = parse_index(simplepy,
             '<a href="svn://svn.zope.org/repos/main/'
             'zope.sqlalchemy/trunk#egg=zope.sqlalchemy-dev" />'
@@ -76,7 +82,7 @@ class TestIndexParsing:
         #assert 0, (result.releaselinks, result.egglinks)
 
     def test_parse_index_normalized_name(self):
-        simplepy = URL("http://pypi.python.org/simple/ndg-httpsclient/")
+        simplepy = URL("https://pypi.org/simple/ndg-httpsclient/")
         result = parse_index(simplepy, """
                <a href="../../pkg/ndg_httpsclient-1.0.tar.gz" />
         """)
@@ -84,7 +90,7 @@ class TestIndexParsing:
         assert result.releaselinks[0].url.endswith("ndg_httpsclient-1.0.tar.gz")
 
     def test_parse_index_two_eggs_same_url(self):
-        simplepy = URL("http://pypi.python.org/simple/Py/")
+        simplepy = URL("https://pypi.org/simple/Py/")
         result = parse_index(simplepy,
             """<a href="../../pkg/pyzip#egg=py-dev">qwe2</a>
                <a href="../../pkg/pyzip#egg=py-dev">qwe</a>
@@ -113,9 +119,10 @@ class TestIndexParsing:
 
     def test_parse_index_invalid_link(self, pypistage):
         result = parse_index(self.simplepy, '''
-                <a rel="download" href="http:/host.com/123" />
+                <a rel="download" href="https:/host.com/123" />
         ''')
-        assert result.crawllinks
+        assert result.crawllinks == {
+            URL('https://pypi.org/host.com/123')}
 
     def test_parse_index_with_egg(self):
         result = parse_index(self.simplepy,
@@ -137,6 +144,37 @@ class TestIndexParsing:
         link, = result.releaselinks
         assert link.basename == "py-1.0-cp27-none-linux_x86_64.whl"
 
+    def test_parse_index_with_requires_python(self):
+        result = parse_index(self.simplepy,
+            """<a href="pkg/py-1.0.zip" data-requires-python="&lt;3" />
+        """)
+        assert len(result.releaselinks) == 1
+        link, = result.releaselinks
+        assert link.basename == "py-1.0.zip"
+        assert link.requires_python == "<3"
+
+    def test_parse_index_with_requires_python_hash_spec_is_better(self):
+        result = parse_index(self.simplepy,
+            """<a href="pkg/py-1.0.zip" data-requires-python="&lt;3" />
+               <a href="pkg/py-1.0.zip#md5=pony"/>
+        """)
+        assert len(result.releaselinks) == 1
+        link, = result.releaselinks
+        assert link.basename == "py-1.0.zip"
+        assert link.hash_spec == "md5=pony"
+        assert link.requires_python is None
+
+    def test_parse_index_with_requires_python_first_with_hash_spec_kept(self):
+        result = parse_index(self.simplepy,
+            """<a href="pkg/py-1.0.zip#md5=pony"/>
+               <a href="pkg/py-1.0.zip#md5=pony" data-requires-python="&lt;3" />
+        """)
+        assert len(result.releaselinks) == 1
+        link, = result.releaselinks
+        assert link.basename == "py-1.0.zip"
+        assert link.hash_spec == "md5=pony"
+        assert link.requires_python is None
+
     @pytest.mark.parametrize("basename", [
         "py-1.3.1.tar.gz",
         "py-1.3.1-1.fc12.src.rpm",
@@ -155,13 +193,13 @@ class TestIndexParsing:
         assert link.basename == basename
 
     def test_parse_index_with_num_in_project(self):
-        simple = URL("http://pypi.python.org/simple/py-4chan/")
+        simple = URL("https://pypi.org/simple/py-4chan/")
         result = parse_index(simple, '<a href="pkg/py-4chan-1.0.zip"/>')
         assert len(result.releaselinks) == 1
         assert result.releaselinks[0].basename == "py-4chan-1.0.zip"
 
     def test_parse_index_unparseable_url(self):
-        simple = URL("http://pypi.python.org/simple/x123/")
+        simple = URL("https://pypi.org/simple/x123/")
         result = parse_index(simple, '<a href="http:" />')
         assert len(result.releaselinks) == 0
 
@@ -219,8 +257,7 @@ class TestIndexParsing:
         assert len(result.crawllinks) == 2
         assert len(result.releaselinks) == 2
         links = list(result.releaselinks)
-        assert links[0].url == \
-                "http://pypi.python.org/pkg/py-1.4.12.zip#md5=12ab"
+        assert links[0].url == "https://pypi.org/pkg/py-1.4.12.zip#md5=12ab"
         assert links[1].url == "http://pylib.org/py-1.1-py27.egg"
 
     def test_releasefile_and_scrape_no_ftp(self):
@@ -252,16 +289,13 @@ class TestIndexParsing:
         assert len(result.crawllinks) == 2
         assert len(result.releaselinks) == 3
         link1, link2, link3 = result.releaselinks
-        assert link1.url == \
-                "http://pypi.python.org/pkg/py-1.4.12.zip#md5=12ab"
-        assert link2.url == \
-                "http://pylib.org/py-1.4.11.zip#md5=1111"
-        assert link3.url == \
-                "http://pypi.python.org/pkg/py-1.4.10.zip#md5=2222"
+        assert link1.url == "https://pypi.org/pkg/py-1.4.12.zip#md5=12ab"
+        assert link2.url == "http://pylib.org/py-1.4.11.zip#md5=1111"
+        assert link3.url == "https://pypi.org/pkg/py-1.4.10.zip#md5=2222"
 
 def test_get_updated(pypistage):
-    c = pypistage.cache_link_updates
-    c2 = pypistage.cache_link_updates
+    c = pypistage.cache_retrieve_times
+    c2 = pypistage.cache_retrieve_times
     return c == c2
 
 class TestExtPYPIDB:
@@ -270,7 +304,7 @@ class TestExtPYPIDB:
         links = pypistage.get_releaselinks("pytest")
         link, = links
         assert link.version == "1.0"
-        assert link.entry.url == "https://pypi.python.org/pytest/pytest-1.0.zip"
+        assert link.entry.url == "https://pypi.org/pytest/pytest-1.0.zip"
         assert not link.hash_spec
         assert link.entrypath.endswith("/pytest-1.0.zip")
         assert link.entrypath == link.entry.relpath
@@ -281,7 +315,7 @@ class TestExtPYPIDB:
         links = pypistage.get_releaselinks("pytest")
         assert links[0].eggfragment == "pytest-dev1"
 
-        pypistage.cache_link_updates.expire("pytest")
+        pypistage.cache_retrieve_times.expire("pytest")
         pypistage.mock_simple("pytest", pypiserial=11,
             pkgver="pytest-1.0.zip#egg=pytest-dev2")
         threadlog.info("hello")
@@ -350,7 +384,7 @@ class TestExtPYPIDB:
             '''.format(md5=md5, hashdir_b=hashdir_b), pypiserial=25)
         links = pypistage.get_releaselinks("pytest")
         assert len(links) == 3
-        assert links[1].entry.url == "https://pypi.python.org/pkg/pytest-1.0.1.zip"
+        assert links[1].entry.url == "https://pypi.org/pkg/pytest-1.0.1.zip"
         assert links[1].entrypath.endswith("/pytest-1.0.1.zip")
 
     def test_parse_and_scrape_non_html_ignored(self, pypistage):
@@ -390,7 +424,7 @@ class TestExtPYPIDB:
         pypistage.keyfs.commit_transaction_in_thread()
         pypistage.keyfs.begin_transaction_in_thread()
         # pretend the last mirror check is very old
-        pypistage.cache_link_updates.expire("pytest")
+        pypistage.cache_retrieve_times.expire("pytest")
 
         # now make sure that we don't cause writes
         commit_serial = pypistage.keyfs.get_current_serial()
@@ -409,7 +443,7 @@ class TestExtPYPIDB:
         links = pypistage.get_releaselinks("pytest")
         assert len(links) == 1
         assert links[0].entry.url == \
-                "https://pypi.python.org/pkg/pytest-1.0.zip"
+                "https://pypi.org/pkg/pytest-1.0.zip"
 
     def test_scrape_not_recursive(self, pypistage):
         pypistage.mock_simple("pytest", text='''
@@ -434,7 +468,7 @@ class TestExtPYPIDB:
         links = pypistage.get_releaselinks("pytest")
         assert len(links) == 1
 
-        # make pypi.python.org unreachable
+        # make pypi.org unreachable
         pypistage.mock_simple("pytest", status_code=-1)
         links2 = pypistage.get_releaselinks("pytest")
         assert links2[0].linkdict == links[0].linkdict and len(links2) == 1
@@ -442,14 +476,14 @@ class TestExtPYPIDB:
         assert len(recs) == 1
 
     def test_pypi_mirror_redirect_to_canonical_issue139(self, pypistage):
-        # GET http://pypi.python.org/simple/Hello_World
+        # GET https://pypi.org/simple/Hello_World
         # will result in the request response to have a "real" URL of
-        # http://pypi.python.org/simple/hello-world because of the
+        # https://pypi.org/simple/hello-world because of the
         # new pypi normalization code
         pypistage.httpget.mock_simple("hello-world",
                 '<a href="Hello_World-1.0.tar.gz" /a>',
                 code=200,
-                url="http://pypi.python.org/simple/hello-world",)
+                url="https://pypi.org/simple/hello-world",)
         l = pypistage.get_releaselinks("Hello_World")
         assert len(l) == 1
 
@@ -458,6 +492,40 @@ class TestExtPYPIDB:
         pypistage.mock_simple("foo", text='<a href="foo-1.0.tar.gz"</a>')
         assert pypistage.has_project_perstage("foo")
 
+    def test_requires_python_caching(self, pypistage):
+        pypistage.mock_simple("foo", text='<a href="foo-1.0.tar.gz" data-requires-python="&lt;3"></a>')
+        (link,) = pypistage.get_releaselinks("foo")
+        assert link.require_python == '<3'
+        # make sure we get the cached data, if not throw an error
+        pypistage.httpget = None
+        (link,) = pypistage.get_releaselinks("foo")
+        assert link.require_python == '<3'
+
+    @pytest.mark.nomocking
+    @pytest.mark.notransaction
+    def test_offline_requires_python(self, mapp, simpypi, testapp, xom):
+        mapp.login('root')
+        mapp.modify_index("root/pypi", indexconfig=dict(type='mirror', mirror_url=simpypi.simpleurl))
+        # turn off offline mode for preparations
+        xom.config.args.offline_mode = False
+        content = b'13'
+        simpypi.add_release('pkg', pkgver='pkg-0.5.zip', requires_python='<3')
+        simpypi.add_release('pkg', pkgver='pkg-1.0.zip')
+        simpypi.add_file('/pkg/pkg-0.5.zip', content)
+        r = testapp.xget(200, '/root/pypi/+simple/pkg/')
+        (link0, link1) = getlinks(r.text)
+        assert link0.text == 'pkg-1.0.zip'
+        assert link0.get('data-requires-python') is None
+        assert link1.text == 'pkg-0.5.zip'
+        assert link1.get('data-requires-python') == '<3'
+        testapp.xget(200, link1['href'].replace('../../', '/root/pypi/'))
+        # turn on offline mode for test
+        xom.config.args.offline_mode = True
+        r = testapp.get('/root/pypi/+simple/pkg/')
+        (link,) = getlinks(r.text)
+        assert link.text == 'pkg-0.5.zip'
+        assert link.get('data-requires-python') == '<3'
+
 
 @pytest.mark.nomockprojectsremote
 class TestPyPIStageprojects:
@@ -483,7 +551,7 @@ class TestPyPIStageprojects:
         assert pypistage.list_projects_perstage() == set(["django"])
         pypistage.mock_simple("proj1", pkgver="proj1-1.0.zip")
         pypistage.mock_simple("proj2", pkgver="proj2-1.0.zip")
-        pypistage.url2response["https://pypi.python.org/simple/proj3/"] = dict(
+        pypistage.url2response["https://pypi.org/simple/proj3/"] = dict(
             status_code=404)
         assert len(pypistage.get_releaselinks("proj1")) == 1
         assert len(pypistage.get_releaselinks("proj2")) == 1
@@ -498,7 +566,6 @@ def raise_ValueError():
 
 @pytest.mark.nomocking
 def test_requests_httpget_negative_status_code(xom, monkeypatch):
-    import requests.exceptions
     l = []
     def r(*a, **k):
         l.append(1)
@@ -509,7 +576,6 @@ def test_requests_httpget_negative_status_code(xom, monkeypatch):
 
 @pytest.mark.nomocking
 def test_requests_httpget_timeout(xom, monkeypatch):
-    import requests.exceptions
     def httpget(url, **kw):
         assert kw["timeout"] == 1.2
         raise requests.exceptions.Timeout()
@@ -520,10 +586,23 @@ def test_requests_httpget_timeout(xom, monkeypatch):
     assert r.status_code == -1
 
 
+@pytest.mark.nomocking
+@pytest.mark.parametrize("exc", [
+    OSError,
+    requests.exceptions.ConnectionError])
+def test_requests_httpget_error(exc, xom, monkeypatch):
+    def httpget(url, **kw):
+        raise exc()
+
+    monkeypatch.setattr(xom._httpsession, "get", httpget)
+    r = xom.httpget("http://notexists.qwe", allow_redirects=False)
+    assert r.status_code == -1
+
+
 def test_is_project_cached(httpget, pypistage):
     assert not pypistage.is_project_cached("xyz")
     assert not pypistage.has_project("xyz")
-    assert pypistage.is_project_cached("xyz")
+    assert not pypistage.is_project_cached("xyz")
 
     httpget.mock_simple("abc", text="")
     assert not pypistage.is_project_cached("abc")
@@ -532,7 +611,9 @@ def test_is_project_cached(httpget, pypistage):
 
 
 def test_404_on_pypi_cached(httpget, pypistage):
-    retrieve_times = pypistage.cache_link_updates
+    # remember current serial to check later
+    serial = pypistage.keyfs.get_current_serial()
+    retrieve_times = pypistage.cache_retrieve_times
     retrieve_times.expire('foo')
     assert not pypistage.has_project_perstage("foo")
     updated_at = retrieve_times.get_timestamp("foo")
@@ -545,10 +626,13 @@ def test_404_on_pypi_cached(httpget, pypistage):
     pypistage.keyfs.commit_transaction_in_thread()
     pypistage.keyfs.begin_transaction_in_thread()
 
-    # we trigger a fresh check and verify that no new commit takes place
-    serial = pypistage.keyfs.get_current_serial()
+    # we trigger a fresh check
     retrieve_times.expire('foo')
     assert not pypistage.has_project_perstage("foo")
+
+    # verify that no new commit took place
+    pypistage.keyfs.commit_transaction_in_thread()
+    pypistage.keyfs.begin_transaction_in_thread()
     assert serial == pypistage.keyfs.get_current_serial()
     updated_at = retrieve_times.get_timestamp('foo')
     assert updated_at > 0
@@ -570,6 +654,10 @@ def test_404_on_pypi_cached(httpget, pypistage):
     assert len(pypistage.get_releaselinks('foo')) == 0
     assert retrieve_times.get_timestamp('foo') > updated_at
 
+    # verify that there was a write this time
+    pypistage.keyfs.commit_transaction_in_thread()
+    assert pypistage.keyfs.get_current_serial() == (serial + 1)
+
 
 class TestProjectNamesCache:
     @pytest.fixture
@@ -586,31 +674,67 @@ class TestProjectNamesCache:
         s2.add(5)
         assert 5 in cache.get()
 
-    def test_is_fresh(self, cache, monkeypatch):
+    def test_is_expired(self, cache, monkeypatch):
         expiry_time = 100
         s = set([1,2,3])
         cache.set(s)
-        assert cache.is_fresh(expiry_time)
+        assert not cache.is_expired(expiry_time)
         t = time.time() + expiry_time + 1
         monkeypatch.setattr("time.time", lambda: t)
-        assert not cache.is_fresh(expiry_time)
+        assert cache.is_expired(expiry_time)
         assert cache.get() == s
 
 
 def test_ProjectUpdateCache(monkeypatch):
     x = ProjectUpdateCache()
     expiry_time = 30
-    assert not x.is_fresh("x", expiry_time)
+    assert x.is_expired("x", expiry_time)
     x.refresh("x")
-    assert x.is_fresh("x", expiry_time)
+    assert not x.is_expired("x", expiry_time)
     t = time.time() + 35
     monkeypatch.setattr("time.time", lambda: t)
-    assert not x.is_fresh("x", expiry_time)
+    assert x.is_expired("x", expiry_time)
     x.refresh("x")
-    assert x.is_fresh("x", expiry_time)
+    assert not x.is_expired("x", expiry_time)
     x.expire("x")
-    assert not x.is_fresh("x", expiry_time)
+    assert x.is_expired("x", expiry_time)
 
     x.refresh("y")
     assert x.get_timestamp("y") == t
 
+
+@pytest.mark.notransaction
+@pytest.mark.with_notifier
+@pytest.mark.nomocking
+def test_redownload_locally_removed_release(mapp, simpypi):
+    from devpi_common.url import URL
+    mapp.create_and_login_user('mirror')
+    indexconfig = dict(
+        type="mirror",
+        mirror_url=simpypi.simpleurl,
+        mirror_cache_expiry=0)
+    mapp.create_index("mirror", indexconfig=indexconfig)
+    mapp.use("mirror/mirror")
+    content = b'14'
+    simpypi.add_release('pkg', pkgver='pkg-1.0.zip')
+    simpypi.add_file('/pkg/pkg-1.0.zip', content)
+    result = mapp.getreleaseslist("pkg")
+    file_relpath = '+files' + URL(result[0]).path
+    assert len(result) == 1
+    r = mapp.downloadrelease(200, result[0])
+    assert r == content
+    with mapp.xom.keyfs.transaction(write=False) as tx:
+        assert tx.conn.io_file_exists(file_relpath)
+    # now remove the local copy
+    with mapp.xom.keyfs.transaction(write=True) as tx:
+        tx.conn.io_file_delete(file_relpath)
+    with mapp.xom.keyfs.transaction(write=False) as tx:
+        assert not tx.conn.io_file_exists(file_relpath)
+    serial = mapp.xom.keyfs.get_current_serial()
+    # and download again
+    r = mapp.downloadrelease(200, result[0])
+    assert r == content
+    with mapp.xom.keyfs.transaction(write=False) as tx:
+        assert tx.conn.io_file_exists(file_relpath)
+    # the serial should not have increased
+    assert serial == mapp.xom.keyfs.get_current_serial()
