@@ -1,5 +1,11 @@
 import pytest
-from devpi_server.main import *
+from devpi_server import mythread
+from devpi_server.config import hookimpl
+from devpi_server.main import Fatal
+from devpi_server.main import XOM
+from devpi_server.main import check_compatible_version
+from devpi_server.main import main
+from devpi_server.main import tween_request_profiling
 import devpi_server
 import os
 
@@ -55,6 +61,7 @@ def test_check_incompatible_version_raises(xom):
 def test_pyramid_configure_called(makexom):
     l = []
     class Plugin:
+        @hookimpl
         def devpiserver_pyramid_configure(self, config, pyramid_config):
             l.append((config, pyramid_config))
     xom = makexom(plugins=[Plugin()])
@@ -79,6 +86,7 @@ def test_run_commands_called(tmpdir):
     from devpi_server.main import _main, get_pluginmanager
     l = []
     class Plugin:
+        @hookimpl
         def devpiserver_cmdline_run(self, xom):
             l.append(xom)
             return 1
@@ -100,6 +108,7 @@ def test_main_starts_server_if_run_commands_returns_none(tmpdir):
     from devpi_server.main import _main, get_pluginmanager
     l = []
     class Plugin:
+        @hookimpl
         def devpiserver_cmdline_run(self, xom):
             l.append(xom)
     pm = get_pluginmanager()
@@ -152,7 +161,7 @@ def test_xom_singleton(xom):
 @pytest.mark.nomocking
 @pytest.mark.parametrize("url", [
     "http://someserver/path",
-    "https://pypi.python.org/simple/package/",
+    "https://pypi.org/simple/package/",
 ])
 @pytest.mark.parametrize("allowRedirect", [True, False])
 def test_offline_mode_httpget_returns_server_error(makexom, url, allowRedirect):
@@ -182,6 +191,22 @@ def test_replica_max_retries_option(makexom, monkeypatch):
                               timeout=1.2)
     assert r.status_code == -1
 
+
+@pytest.mark.nomocking
+@pytest.mark.parametrize("input_set", [
+    {'timeout': 5, 'arg': [], 'kwarg': None},
+    {'timeout': 42, 'arg': ["--request-timeout=42"], 'kwarg': None},
+    {'timeout': 123, 'arg': [], 'kwarg': 123}
+])
+def test_request_args_timeout_handover(makexom, input_set):
+    def mock_http_get(*args, **kwargs):
+        assert kwargs["timeout"] == input_set['timeout']
+    xom = makexom(input_set['arg'])
+    xom._httpsession.get = mock_http_get
+
+    xom.httpget("http://whatever", allow_redirects=False, timeout=input_set['kwarg'])
+
+
 def test_no_root_pypi_option(makexom):
     xom = makexom(["--no-root-pypi"])
     with xom.keyfs.transaction(write=False):
@@ -233,8 +258,16 @@ def test_init_server_directory(call_devpi_in_dir, tmpdir):
 
 
 def test_serve_threads(monkeypatch, tmpdir):
-    def check_threads(app, host, port, threads):
+    def check_threads(app, host, port, threads, max_request_body_size):
         assert threads == 100
     monkeypatch.setattr("waitress.serve", check_threads)
     from devpi_server.main import main
     main(["devpi-server", "--threads", "100"])
+
+
+def test_serve_max_body(monkeypatch, tmpdir):
+    def check_max_body(app, host, port, threads, max_request_body_size):
+        assert max_request_body_size == 42
+    monkeypatch.setattr("waitress.serve", check_max_body)
+    from devpi_server.main import main
+    main(["devpi-server", "--max-request-body-size", "42"])
