@@ -1,12 +1,15 @@
 from __future__ import unicode_literals
-
+import getpass
 import py
 import pytest
 import json
 
 from devpi_common.metadata import splitbasename
 from devpi_common.archive import Archive, zip_dict
-from devpi_server.model import *  # noqa
+from devpi_server.config import hookimpl
+from devpi_server.model import InvalidIndexconfig
+from devpi_server.model import get_indexconfig
+from devpi_server.model import run_passwd
 from py.io import BytesIO
 
 pytestmark = [pytest.mark.writetransaction]
@@ -753,6 +756,7 @@ class TestStage:
     @pytest.mark.start_threads
     def test_set_versiondata_hook(self, stage, queue):
         class Plugin:
+            @hookimpl
             def devpiserver_on_changed_versiondata(self,
                     stage, project, version, metadata):
                 queue.put((stage, metadata))
@@ -773,6 +777,7 @@ class TestStage:
     @pytest.mark.notransaction
     def test_stage_created_hook(self, xom, queue):
         class Plugin:
+            @hookimpl
             def devpiserver_stage_created(self, stage):
                 queue.put(stage)
         xom.config.pluginmanager.register(Plugin())
@@ -790,10 +795,10 @@ class TestStage:
             user.create_stage("hello2")
         assert queue.get(timeout=10).name == "user/hello2"
 
-
     @pytest.mark.start_threads
     def test_doczip_uploaded_hook(self, stage, queue):
         class Plugin:
+            @hookimpl
             def devpiserver_on_upload(self, stage, project, version, link):
                 queue.put((stage, project, version, link))
         stage.xom.config.pluginmanager.register(Plugin())
@@ -820,6 +825,38 @@ class TestStage:
         with stage.xom.keyfs.transaction():
             assert link.entry.file_exists()
 
+    @pytest.mark.start_threads
+    def test_doczip_remove_hook(self, stage, queue):
+        class Plugin:
+            @hookimpl
+            def devpiserver_on_upload(self, stage, project, version, link):
+                queue.put((stage, project, version, link))
+        stage.xom.config.pluginmanager.register(Plugin())
+        class Plugin:
+            @hookimpl
+            def devpiserver_on_remove_file(self, stage, relpath):
+                queue.put((stage, relpath))
+        stage.xom.config.pluginmanager.register(Plugin())
+
+        # upload, should trigger devpiserver_on_upload
+        stage.set_versiondata(udict(name="pkg2", version="1.0"))
+        content = zip_dict({"index.html": "<html/>",
+            "_static": {}, "_templ": {"x.css": ""}})
+        stage.store_doczip("pkg2", "1.0", content)
+        stage.xom.keyfs.commit_transaction_in_thread()
+        nstage, name, version, link = queue.get()
+        assert name == "pkg2"
+        assert version == "1.0"
+        with stage.xom.keyfs.transaction():
+            assert link.entry.file_get_content() == content
+
+        # remove, should trigger devpiserver_on_remove_file
+        with stage.xom.keyfs.transaction(write=True):
+            linkstore = stage.get_linkstore_perstage("pkg2", "1.0", readonly=False)
+            linkstore.remove_links()
+        nstage, relpath = queue.get()
+        assert relpath.startswith('hello/world/+')
+        assert relpath.endswith('/pkg2-1.0.doc.zip')
 
     def test_get_existing_project(self, stage):
         assert not stage.get_versiondata("hello", "1.0")
@@ -935,7 +972,7 @@ class TestUsers:
         assert newusers == set("user1 user2".split())
 
     def test_server_passwd(self, model, monkeypatch):
-        monkeypatch.setattr(py.std.getpass, "getpass", lambda x: "123")
+        monkeypatch.setattr(getpass, "getpass", lambda x: "123")
         run_passwd(model, "root")
         assert model.get_user("root").validate("123")
 
@@ -952,7 +989,7 @@ class TestUsers:
         userconfig = user.get(credentials=True)
         assert userconfig['pwsalt'] is not None
         assert userconfig['pwhash'] is not None
-        monkeypatch.setattr(py.std.getpass, "getpass", lambda x: new_secret)
+        monkeypatch.setattr(getpass, "getpass", lambda x: new_secret)
         run_passwd(model, "root")
         assert model.get_user("root").validate(new_secret)
 
@@ -989,6 +1026,7 @@ def test_setdefault_indexes(xom, model):
     ("", []), ("x,y", ["x", "y"]), ("x,,y", ["x", "y"])))
 def test_get_indexconfig_lists(key, value, result):
     class hooks:
+        @hookimpl
         def devpiserver_indexconfig_defaults(self, index_type):
             return {}
     kvdict = get_indexconfig(hooks(), type="stage", **{key: value})
@@ -1024,6 +1062,7 @@ def test_get_indexconfig_lists(key, value, result):
 ])
 def test_get_indexconfig_values(xom, input, expected):
     class hooks:
+        @hookimpl
         def devpiserver_indexconfig_defaults(self, index_type):
             return {}
     result = get_indexconfig(hooks(), type="stage", **input)
