@@ -1,16 +1,33 @@
 # -*- coding: utf-8 -*-
 import hashlib
+import json
 import pytest
 import py
 from devpi_server.log import thread_pop_log
 from devpi_server.fileutil import load
-from devpi_server.replica import *  # noqa
+from devpi_server.log import threadlog, thread_push_log
+from devpi_server.replica import H_EXPECTED_MASTER_ID, H_MASTER_UUID
+from devpi_server.replica import H_REPLICA_UUID, H_REPLICA_OUTSIDE_URL
+from devpi_server.replica import FileReplicationError
+from devpi_server.replica import ImportFileReplica
+from devpi_server.replica import MasterChangelogRequest
+from devpi_server.replica import ReplicaThread
+from devpi_server.replica import tween_replica_proxy
+from pyramid.httpexceptions import HTTPNotFound
+from pyramid.response import Response
+
 
 def loads(bytestring):
     return load(py.io.BytesIO(bytestring))
 
 pytestmark = [pytest.mark.notransaction]
 
+
+@pytest.fixture
+def replica_pypistage(devpiserver_makepypistage, replica_xom):
+    return devpiserver_makepypistage(replica_xom)
+
+
 @pytest.fixture
 def testapp(testapp):
     testapp.xom.config.nodeinfo["role"] = "master"
@@ -143,7 +160,7 @@ class TestReplicaThread:
         # run
         with pytest.raises(ZeroDivisionError):
             rt.thread_run()
-        msgs = [x.msg for x in caplog.getrecords(".*http://localhost/\+changelog/0")]
+        msgs = [x.msg for x in caplog.getrecords(r".*http://localhost/\+changelog/0")]
         assert msgs == [
             '[REP] fetching %s',
             '[REP] error fetching %s: %s',
@@ -389,7 +406,7 @@ class TestFileReplication:
         md5 = hashlib.md5(content1).hexdigest()
         link = gen.pypi_package_link("pytest-1.8.zip#md5=%s" % md5, md5=False)
         with xom.keyfs.transaction(write=True):
-            entry = xom.filestore.maplink(link, "root", "pypi")
+            entry = xom.filestore.maplink(link, "root", "pypi", "pytest")
             assert not entry.file_exists()
 
         replay(xom, replica_xom)
@@ -441,7 +458,7 @@ class TestFileReplication:
         md5 = hashlib.md5(content1).hexdigest()
         link = gen.pypi_package_link("pytest-1.8.zip#md5=%s" % md5, md5=False)
         with xom.keyfs.transaction(write=True):
-            entry = xom.filestore.maplink(link, "root", "pypi")
+            entry = xom.filestore.maplink(link, "root", "pypi", "pytest")
             assert not entry.file_exists()
 
         master_url = replica_xom.config.master_url
@@ -472,7 +489,7 @@ class TestFileReplication:
         content1 = b'hello'
         link = gen.pypi_package_link("some-1.8.zip", md5=False)
         with xom.keyfs.transaction(write=True):
-            entry = xom.filestore.maplink(link, "root", "pypi")
+            entry = xom.filestore.maplink(link, "root", "pypi", "some")
             assert not entry.file_exists()
             assert not entry.hash_spec
 
@@ -493,7 +510,7 @@ class TestFileReplication:
                                          content=b'')
         with pytest.raises(FileReplicationError) as e:
             replay(xom, replica_xom)
-        assert str(e.value) == 'FileReplicationError with http://localhost/root/pypi/+e/https_pypi.python.org_package_some/some-1.8.zip, code=500, relpath=root/pypi/+e/https_pypi.python.org_package_some/some-1.8.zip, message=failed'
+        assert str(e.value) == 'FileReplicationError with http://localhost/root/pypi/+e/https_pypi.org_package_some/some-1.8.zip, code=500, reason=Internal Server Error, relpath=root/pypi/+e/https_pypi.org_package_some/some-1.8.zip, message=failed'
 
         # now get the real thing
         replica_xom.httpget.mockresponse(master_file_path, status_code=200,
@@ -510,7 +527,7 @@ class TestFileReplication:
                             lambda x: l.append(x))
         with xom.keyfs.transaction(write=True):
             link = gen.pypi_package_link("pytest-1.8.zip", md5=True)
-            entry = xom.filestore.maplink(link, "root", "pypi")
+            entry = xom.filestore.maplink(link, "root", "pypi", "pytest")
             assert entry.hash_spec and not entry.file_exists()
         replay(xom, replica_xom)
         with replica_xom.keyfs.transaction():
@@ -521,7 +538,7 @@ class TestFileReplication:
                 for part in entry.iter_remote_file_replica():
                     pass
             e.match('received 500 from master')
-            e.match('pypi.python.org/package/some/pytest-1.8.zip: received 404')
+            e.match('pypi.org/package/some/pytest-1.8.zip: received 404')
 
     def test_cache_remote_file_fetch_original(self, xom, replica_xom, gen,
                                               pypistage, monkeypatch, reqmock):
@@ -529,11 +546,11 @@ class TestFileReplication:
         monkeypatch.setattr(xom.keyfs, "wait_tx_serial",
                             lambda x: l.append(x))
         with xom.keyfs.transaction(write=True):
-            md5 = py.std.hashlib.md5()
+            md5 = hashlib.md5()
             md5.update(b'123')
             link = gen.pypi_package_link(
                 "pytest-1.8.zip", md5=md5.hexdigest())
-            entry = xom.filestore.maplink(link, "root", "pypi")
+            entry = xom.filestore.maplink(link, "root", "pypi", "pytest")
             assert entry.hash_spec and not entry.file_exists()
         replay(xom, replica_xom)
         with replica_xom.keyfs.transaction():
@@ -566,6 +583,7 @@ class TestFileReplication:
         # first we try to return something wrong
         master_url = replica_xom.config.master_url
         (path,) = mapp.get_release_paths('hello')
+        file_relpath = '+files' + path
         master_file_url = master_url.joinpath(path).url
         replica_xom.httpget.mockresponse(master_file_url, code=200, content=b'13')
         replay(xom, replica_xom)
@@ -581,9 +599,13 @@ class TestFileReplication:
                    "X-DEVPI-SERIAL": str(xom.keyfs.get_current_serial())}
         reqmock.mockresponse(master_file_url, code=200, headers=headers)
         replica_xom.httpget.mockresponse(master_file_url, code=200, content=content1, headers=headers)
+        with replica_xom.keyfs.transaction(write=False) as tx:
+            assert not tx.conn.io_file_exists(file_relpath)
         r = r_app.get(path)
         assert r.status_code == 200
         assert r.body == content1
+        with replica_xom.keyfs.transaction(write=False) as tx:
+            assert tx.conn.io_file_exists(file_relpath)
         replication_errors = ReplicationErrors(replica_xom.config.serverdir)
         assert list(replication_errors.errors.keys()) == []
 
@@ -636,17 +658,17 @@ def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypis
     serial = xom.keyfs.get_current_serial()
     httpget.mock_simple(
         'pytest',
-        text='<a href="https://pypi.python.org/pytest/pytest-1.0.zip">pytest-1.0.zip</a>',
+        text='<a href="https://pypi.org/pytest/pytest-1.0.zip">pytest-1.0.zip</a>',
         headers={'X-DEVPI-SERIAL': str(serial)})
     with replica_xom.keyfs.transaction():
         ret = replica_pypistage.get_releaselinks("pytest")
     assert len(ret) == 1
-    assert ret[0].relpath == 'root/pypi/+e/https_pypi.python.org_pytest/pytest-1.0.zip'
+    assert ret[0].relpath == 'root/pypi/+e/https_pypi.org_pytest/pytest-1.0.zip'
 
     # now we change the links and expire the cache
     pypiurls.simple = orig_simple
     pypistage.mock_simple("pytest", pkgver="pytest-1.1.zip", pypiserial=10001)
-    pypistage.cache_link_updates.expire('pytest')
+    pypistage.cache_retrieve_times.expire('pytest')
     with xom.keyfs.transaction(write=True):
         pypistage.get_releaselinks("pytest")
     assert xom.keyfs.get_current_serial() > serial
@@ -664,18 +686,18 @@ def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypis
     pypiurls.simple = 'http://localhost:3111/root/pypi/+simple/'
     httpget.mock_simple(
         'pytest',
-        text='<a href="https://pypi.python.org/pkg/pytest-1.1.zip">pytest-1.1.zip</a>',
+        text='<a href="https://pypi.org/pkg/pytest-1.1.zip">pytest-1.1.zip</a>',
         pypiserial=10001,
         headers={'X-DEVPI-SERIAL': str(xom.keyfs.get_current_serial())})
     with replica_xom.keyfs.transaction():
         # make the replica believe it hasn't updated for a longer time
         r_pypistage = replica_xom.model.getstage("root/pypi")
-        r_pypistage.cache_link_updates.expire("pytest")
+        r_pypistage.cache_retrieve_times.expire("pytest")
         ret = replica_pypistage.get_releaselinks("pytest")
     assert called == [True]
     replay(xom, replica_xom)
     assert len(ret) == 1
-    assert ret[0].relpath == 'root/pypi/+e/https_pypi.python.org_pytest/pytest-1.1.zip'
+    assert ret[0].relpath == 'root/pypi/+e/https_pypi.org_pytest/pytest-1.1.zip'
 
 
 def test_replicate_deleted_user(mapp, replica_xom):
