@@ -1,7 +1,15 @@
 import pytest
+import sys
 from . import test_streaming
 
 
+pytestmark = [
+    pytest.mark.skipif(
+        sys.platform == "win32",
+        reason="issues with process management on Windows"),
+    pytest.mark.skipif("not config.option.slow")]
+
+
 @pytest.fixture
 def host_port(replica_host_port):
     return replica_host_port
